import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.resizing import AddLinesModule
from pynpoint.processing.timedenoising import (
    CwtWaveletConfiguration,
    DwtWaveletConfiguration,
    WaveletTimeDenoisingModule,
    TimeNormalizationModule,
)
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestTimeDenoising:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "images")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["images"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read",
            image_tag="images",
            input_dir=self.test_dir + "images",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read")

        data = self.pipeline.get_data("images")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_wavelet_denoising_cwt_dog(self) -> None:

        cwt_config = CwtWaveletConfiguration(
            wavelet="dog", wavelet_order=2, keep_mean=False, resolution=0.5
        )

        assert cwt_config.m_wavelet == "dog"
        assert cwt_config.m_wavelet_order == 2
        assert not cwt_config.m_keep_mean
        assert cwt_config.m_resolution == 0.5

        module = WaveletTimeDenoisingModule(
            wavelet_configuration=cwt_config,
            name_in="wavelet_cwt_dog",
            image_in_tag="images",
            image_out_tag="wavelet_cwt_dog",
            padding="zero",
            median_filter=True,
            threshold_function="soft",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("wavelet_cwt_dog")

        data = self.pipeline.get_data("wavelet_cwt_dog")
        assert np.sum(data) == pytest.approx(105.1035789572968, rel=self.limit, abs=0.0)
        assert data.shape == (10, 11, 11)

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 4

        self.pipeline.run_module("wavelet_cwt_dog")

        data_multi = self.pipeline.get_data("wavelet_cwt_dog")
        assert data == pytest.approx(data_multi, rel=self.limit, abs=0.0)
        assert data.shape == data_multi.shape

    def test_wavelet_denoising_cwt_morlet(self) -> None:

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 1

        cwt_config = CwtWaveletConfiguration(
            wavelet="morlet", wavelet_order=5, keep_mean=False, resolution=0.5
        )

        assert cwt_config.m_wavelet == "morlet"
        assert cwt_config.m_wavelet_order == 5
        assert not cwt_config.m_keep_mean
        assert cwt_config.m_resolution == 0.5

        module = WaveletTimeDenoisingModule(
            wavelet_configuration=cwt_config,
            name_in="wavelet_cwt_morlet",
            image_in_tag="images",
            image_out_tag="wavelet_cwt_morlet",
            padding="mirror",
            median_filter=False,
            threshold_function="hard",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("wavelet_cwt_morlet")

        data = self.pipeline.get_data("wavelet_cwt_morlet")
        assert np.sum(data) == pytest.approx(
            104.86262840716438, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_attribute(
            "wavelet_cwt_morlet", "NFRAMES", static=False
        )
        assert data[0] == data[1] == 5

    def test_wavelet_denoising_dwt(self) -> None:

        dwt_config = DwtWaveletConfiguration(wavelet="db8")

        assert dwt_config.m_wavelet == "db8"

        module = WaveletTimeDenoisingModule(
            wavelet_configuration=dwt_config,
            name_in="wavelet_dwt",
            image_in_tag="images",
            image_out_tag="wavelet_dwt",
            padding="zero",
            median_filter=True,
            threshold_function="soft",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("wavelet_dwt")

        data = self.pipeline.get_data("wavelet_dwt")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_time_normalization(self) -> None:

        module = TimeNormalizationModule(
            name_in="timenorm", image_in_tag="images", image_out_tag="timenorm"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("timenorm")

        data = self.pipeline.get_data("timenorm")
        assert np.sum(data) == pytest.approx(56.443663773873, rel=self.limit, abs=0.0)
        assert data.shape == (10, 11, 11)

    def test_wavelet_denoising_even_size(self) -> None:

        module = AddLinesModule(
            name_in="add",
            image_in_tag="images",
            image_out_tag="images_even",
            lines=(1, 0, 1, 0),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("add")

        data = self.pipeline.get_data("images_even")
        assert np.sum(data) == pytest.approx(
            105.54278879805275, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 12, 12)

        cwt_config = CwtWaveletConfiguration(
            wavelet="dog", wavelet_order=2, keep_mean=False, resolution=0.5
        )

        assert cwt_config.m_wavelet == "dog"
        assert cwt_config.m_wavelet_order == 2
        assert not cwt_config.m_keep_mean
        assert cwt_config.m_resolution == 0.5

        module = WaveletTimeDenoisingModule(
            wavelet_configuration=cwt_config,
            name_in="wavelet_even_1",
            image_in_tag="images_even",
            image_out_tag="wavelet_even_1",
            padding="zero",
            median_filter=True,
            threshold_function="soft",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("wavelet_even_1")

        data = self.pipeline.get_data("wavelet_even_1")
        assert np.sum(data) == pytest.approx(105.1035789572968, rel=self.limit, abs=0.0)
        assert data.shape == (10, 12, 12)

        module = WaveletTimeDenoisingModule(
            wavelet_configuration=cwt_config,
            name_in="wavelet_even_2",
            image_in_tag="images_even",
            image_out_tag="wavelet_even_2",
            padding="mirror",
            median_filter=True,
            threshold_function="soft",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("wavelet_even_2")

        data = self.pipeline.get_data("wavelet_even_2")
        assert np.sum(data) == pytest.approx(
            105.06809820408587, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 12, 12)

        data = self.pipeline.get_attribute("images", "NFRAMES", static=False)
        assert data == pytest.approx([5, 5], rel=self.limit, abs=0.0)

        data = self.pipeline.get_attribute("wavelet_even_1", "NFRAMES", static=False)
        assert data == pytest.approx([5, 5], rel=self.limit, abs=0.0)

        data = self.pipeline.get_attribute("wavelet_even_2", "NFRAMES", static=False)
        assert data == pytest.approx([5, 5], rel=self.limit, abs=0.0)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.centering import (
    StarAlignmentModule,
    ShiftImagesModule,
    WaffleCenteringModule,
    FitCenterModule,
)
from pynpoint.processing.extract import StarExtractionModule
from pynpoint.processing.resizing import AddLinesModule
from pynpoint.util.tests import (
    create_config,
    create_star_data,
    create_waffle_data,
    remove_test_data,
)


class TestCentering:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "star")
        create_waffle_data(self.test_dir + "waffle")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(path=self.test_dir, folders=["star", "waffle"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read1",
            image_tag="star",
            input_dir=self.test_dir + "star",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)

        module = FitsReadingModule(
            name_in="read2",
            image_tag="waffle",
            input_dir=self.test_dir + "waffle",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)

        self.pipeline.run_module("read1")
        self.pipeline.run_module("read2")

        data = self.pipeline.get_data("star")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_data("waffle")
        assert np.sum(data) == pytest.approx(4.000000000000196, rel=self.limit, abs=0.0)
        assert data.shape == (1, 101, 101)

    def test_star_extract(self) -> None:

        module = StarExtractionModule(
            name_in="extract1",
            image_in_tag="star",
            image_out_tag="extract1",
            index_out_tag="index",
            image_size=0.2,
            fwhm_star=0.1,
            position=None,
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("extract1")

        assert len(warning) == 3

        assert (
            warning[0].message.args[0] == "Can not store the attribute 'INSTRUMENT' "
            "because the dataset 'index' does not exist."
        )

        data = self.pipeline.get_data("extract1")
        assert np.sum(data) == pytest.approx(
            104.93318507061295, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

    def test_star_align(self) -> None:

        module = StarAlignmentModule(
            name_in="align1",
            image_in_tag="extract1",
            ref_image_in_tag=None,
            image_out_tag="align1",
            accuracy=10,
            resize=2.0,
            num_references=10,
            subframe=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("align1")

        data = self.pipeline.get_data("align1")
        assert np.sum(data) == pytest.approx(104.7074742320535, rel=self.limit, abs=0.0)
        assert data.shape == (10, 18, 18)

    def test_star_align_subframe(self) -> None:

        module = StarAlignmentModule(
            name_in="align2",
            image_in_tag="extract1",
            ref_image_in_tag=None,
            image_out_tag="align2",
            accuracy=10,
            resize=None,
            num_references=10,
            subframe=0.1,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("align2")

        data = self.pipeline.get_data("align2")
        assert np.sum(data) == pytest.approx(
            104.39031104541652, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

    def test_star_align_ref(self) -> None:

        module = StarAlignmentModule(
            name_in="align3",
            image_in_tag="extract1",
            ref_image_in_tag="align2",
            image_out_tag="align3",
            accuracy=10,
            resize=None,
            num_references=10,
            subframe=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("align3")

        data = self.pipeline.get_data("align3")
        assert np.sum(data) == pytest.approx(
            104.46997194330757, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

    def test_star_align_number_ref(self) -> None:

        module = StarAlignmentModule(
            name_in="align4",
            image_in_tag="extract1",
            ref_image_in_tag="align2",
            image_out_tag="align4",
            accuracy=10,
            resize=None,
            num_references=20,
            subframe=None,
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("align4")

        assert len(warning) == 1

        assert (
            warning[0].message.args[0]
            == "Number of available images (10) is smaller than "
            "num_references (20). Using all available images "
            "instead."
        )

        data = self.pipeline.get_data("align4")
        assert np.sum(data) == pytest.approx(
            104.46997194330757, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

    def test_shift_images_spline(self) -> None:

        module = ShiftImagesModule(
            shift_xy=(1.0, 2.0),
            interpolation="spline",
            name_in="shift1",
            image_in_tag="align1",
            image_out_tag="shift",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("shift1")

        data = self.pipeline.get_data("shift")
        assert np.sum(data) == pytest.approx(
            104.20425101355244, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 18, 18)

    def test_shift_images_fft(self) -> None:

        module = ShiftImagesModule(
            shift_xy=(1.0, 2.0),
            interpolation="fft",
            name_in="shift2",
            image_in_tag="align1",
            image_out_tag="shift_fft",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("shift2")

        data = self.pipeline.get_data("shift_fft")
        assert np.sum(data) == pytest.approx(104.7074742320535, rel=self.limit, abs=0.0)
        assert data.shape == (10, 18, 18)

    def test_waffle_center_odd(self) -> None:

        module = AddLinesModule(
            name_in="add",
            image_in_tag="star",
            image_out_tag="star_add",
            lines=(45, 45, 45, 45),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("add")

        data = self.pipeline.get_data("star_add")
        assert np.sum(data) == pytest.approx(
            105.54278879805278, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 101, 101)

        module = WaffleCenteringModule(
            size=0.2,
            center=(50, 50),
            name_in="waffle",
            image_in_tag="star_add",
            center_in_tag="waffle",
            image_out_tag="center",
            radius=35.0,
            pattern="x",
            sigma=0.05,
        )

        self.pipeline.add_module(module)

        with pytest.warns(DeprecationWarning) as warning:
            self.pipeline.run_module("waffle")

        assert len(warning) == 1

        assert (
            warning[0].message.args[0]
            == "The 'pattern' parameter will be deprecated in a "
            "future release. Please Use the 'angle' "
            "parameter instead and set it to 45.0 degrees."
        )

        data = self.pipeline.get_data("center")
        assert np.sum(data) == pytest.approx(
            104.93318507061295, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

        attr = self.pipeline.get_attribute("center", "History: WaffleCenteringModule")
        assert attr == "[x, y] = [50.0, 50.0]"

    def test_waffle_center_even(self) -> None:

        module = AddLinesModule(
            name_in="add1",
            image_in_tag="star_add",
            image_out_tag="star_even",
            lines=(0, 1, 0, 1),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("add1")

        data = self.pipeline.get_data("star_even")
        assert np.sum(data) == pytest.approx(
            105.54278879805275, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 102, 102)

        module = AddLinesModule(
            name_in="add2",
            image_in_tag="waffle",
            image_out_tag="waffle_even",
            lines=(0, 1, 0, 1),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("add2")

        data = self.pipeline.get_data("waffle_even")
        assert np.sum(data) == pytest.approx(4.000000000000195, rel=self.limit, abs=0.0)
        assert data.shape == (1, 102, 102)

        module = ShiftImagesModule(
            shift_xy=(0.5, 0.5),
            interpolation="spline",
            name_in="shift3",
            image_in_tag="star_even",
            image_out_tag="star_shift",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("shift3")

        data = self.pipeline.get_data("star_shift")
        assert np.sum(data) == pytest.approx(
            105.54278879805274, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 102, 102)

        module = ShiftImagesModule(
            shift_xy=(0.5, 0.5),
            interpolation="spline",
            name_in="shift4",
            image_in_tag="waffle_even",
            image_out_tag="waffle_shift",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("shift4")

        data = self.pipeline.get_data("waffle_shift")
        assert np.sum(data) == pytest.approx(4.000000000000194, rel=self.limit, abs=0.0)
        assert data.shape == (1, 102, 102)

        module = WaffleCenteringModule(
            size=0.2,
            center=(50, 50),
            name_in="waffle_even",
            image_in_tag="star_shift",
            center_in_tag="waffle_shift",
            image_out_tag="center_even",
            radius=35.0,
            pattern="x",
            sigma=0.05,
        )

        self.pipeline.add_module(module)

        with pytest.warns(DeprecationWarning) as warning:
            self.pipeline.run_module("waffle_even")

        assert len(warning) == 1

        assert (
            warning[0].message.args[0]
            == "The 'pattern' parameter will be deprecated in a "
            "future release. Please Use the 'angle' "
            "parameter instead and set it to 45.0 degrees."
        )

        data = self.pipeline.get_data("center_even")
        assert np.sum(data) == pytest.approx(
            105.22695036281449, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

        attr = self.pipeline.get_attribute(
            "center_even", "History: WaffleCenteringModule"
        )
        assert attr == "[x, y] = [50.5, 50.5]"

    def test_fit_center_full(self) -> None:

        module = FitCenterModule(
            name_in="fit1",
            image_in_tag="shift",
            fit_out_tag="fit_full",
            mask_out_tag="mask",
            method="full",
            mask_radii=(None, 0.1),
            sign="positive",
            model="gaussian",
            guess=(1.0, 2.0, 3.0, 3.0, 0.01, 0.0, 0.0),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("fit1")

        data = self.pipeline.get_data("fit_full")

        assert np.mean(data[:, 0]) == pytest.approx(0.94, rel=0.0, abs=0.05)
        assert np.mean(data[:, 2]) == pytest.approx(2.07, rel=0.0, abs=0.05)
        assert np.mean(data[:, 4]) == pytest.approx(0.0676, rel=0.0, abs=0.05)
        assert np.mean(data[:, 6]) == pytest.approx(0.0665, rel=0.0, abs=0.05)
        assert np.mean(data[:, 8]) == pytest.approx(0.24, rel=0.0, abs=0.05)
        assert data.shape == (10, 14)

        data = self.pipeline.get_data("mask")
        assert np.sum(data) == pytest.approx(
            103.45599730750453, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 18, 18)

    def test_fit_center_mean(self) -> None:

        module = FitCenterModule(
            name_in="fit2",
            image_in_tag="shift",
            fit_out_tag="fit_mean",
            mask_out_tag=None,
            method="mean",
            mask_radii=(None, 0.1),
            sign="positive",
            model="moffat",
            guess=(1.0, 2.0, 3.0, 3.0, 0.01, 0.0, 0.0, 1.0),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("fit2")

        data = self.pipeline.get_data("fit_mean")
        assert np.mean(data[:, 0]) == pytest.approx(0.94, rel=0.0, abs=0.01)
        assert np.mean(data[:, 2]) == pytest.approx(2.059, rel=0.0, abs=0.01)
        assert np.mean(data[:, 4]) == pytest.approx(0.083, rel=0.0, abs=0.01)
        assert np.mean(data[:, 6]) == pytest.approx(0.08, rel=0.0, abs=0.01)
        assert np.mean(data[:, 8]) == pytest.approx(0.242, rel=0.0, abs=0.01)
        assert data.shape == (10, 16)

    def test_shift_images_tag(self) -> None:

        module = ShiftImagesModule(
            shift_xy="fit_full",
            interpolation="spline",
            name_in="shift5",
            image_in_tag="shift",
            image_out_tag="shift_tag_1",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("shift5")

        data = self.pipeline.get_data("shift_tag_1")
        assert np.sum(data) == pytest.approx(103.76410960085425, rel=1e-6, abs=0.9)
        assert data.shape == (10, 18, 18)

    def test_shift_images_tag_mean(self) -> None:

        module = ShiftImagesModule(
            shift_xy="fit_mean",
            interpolation="spline",
            name_in="shift6",
            image_in_tag="shift",
            image_out_tag="shift_tag_2",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("shift6")

        data = self.pipeline.get_data("shift_tag_2")
        assert np.sum(data) == pytest.approx(103.42285439876926, rel=1e-6, abs=0.0)
        assert data.shape == (10, 18, 18)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.psfpreparation import (
    PSFpreparationModule,
    AngleInterpolationModule,
    AngleCalculationModule,
    SDIpreparationModule,
    SortParangModule,
)
from pynpoint.util.tests import (
    create_config,
    create_star_data,
    create_ifs_data,
    remove_test_data,
)


class TestPsfPreparation:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "prep")
        create_ifs_data(self.test_dir + "prep_ifs")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["prep", "prep_ifs"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read", image_tag="read", input_dir=self.test_dir + "prep"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read")

        data = self.pipeline.get_data("read")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(
            name_in="read_ifs",
            image_tag="read_ifs",
            input_dir=self.test_dir + "prep_ifs",
            ifs_data=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read_ifs")

        data = self.pipeline.get_data("read_ifs")
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.0)
        assert data.shape == (3, 10, 21, 21)

    def test_angle_interpolation(self) -> None:

        module = AngleInterpolationModule(name_in="angle1", data_tag="read")

        self.pipeline.add_module(module)
        self.pipeline.run_module("angle1")

        data = self.pipeline.get_data("header_read/PARANG")
        assert np.sum(data) == pytest.approx(900.0, rel=self.limit, abs=0.0)
        assert data.shape == (10,)

    def test_angle_calculation(self) -> None:

        self.pipeline.set_attribute("read", "LATITUDE", -25.0)
        self.pipeline.set_attribute("read", "LONGITUDE", -70.0)
        self.pipeline.set_attribute("read", "DIT", 1.0)

        self.pipeline.set_attribute(
            "read", "RA", (90.0, 90.0, 90.0, 90.0), static=False
        )
        self.pipeline.set_attribute(
            "read", "DEC", (-51.0, -51.0, -51.0, -51.0), static=False
        )
        self.pipeline.set_attribute(
            "read", "PUPIL", (90.0, 90.0, 90.0, 90.0), static=False
        )

        date = (
            "2012-12-01T07:09:00.0000",
            "2012-12-01T07:09:01.0000",
            "2012-12-01T07:09:02.0000",
            "2012-12-01T07:09:03.0000",
        )

        self.pipeline.set_attribute("read", "DATE", date, static=False)

        module = AngleCalculationModule(
            instrument="NACO", name_in="angle2", data_tag="read"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("angle2")

        data = self.pipeline.get_data("header_read/PARANG")
        assert np.sum(data) == pytest.approx(
            -550.2338293743467, rel=self.limit, abs=0.0
        )
        assert data.shape == (10,)

        self.pipeline.set_attribute(
            "read", "RA", (60000.0, 60000.0, 60000.0, 60000.0), static=False
        )

        self.pipeline.set_attribute(
            "read", "DEC", (-510000.0, -510000.0, -510000.0, -510000.0), static=False
        )

        module = AngleCalculationModule(
            instrument="SPHERE/IRDIS", name_in="angle3", data_tag="read"
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("angle3")

        warning_0 = (
            "For SPHERE data it is recommended to use the header keyword 'ESO INS4 "
            "DROT2 RA' to specify the object's right ascension. The input will be "
            "parsed accordingly. Using the regular 'RA' keyword will lead to wrong "
            "parallactic angles."
        )

        warning_1 = (
            "For SPHERE data it is recommended to use the header keyword 'ESO INS4 "
            "DROT2 DEC' to specify the object's declination. The input will be parsed "
            "accordingly. Using the regular 'DEC' keyword will lead to wrong "
            "parallactic angles."
        )

        if len(warning) == 2:
            assert warning[0].message.args[0] == warning_0
            assert warning[1].message.args[0] == warning_1

        data = self.pipeline.get_data("header_read/PARANG")
        assert np.sum(data) == pytest.approx(
            1704.2202367435675, rel=self.limit, abs=0.0
        )
        assert data.shape == (10,)

        module = AngleCalculationModule(
            instrument="SPHERE/IFS", name_in="angle4", data_tag="read"
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("angle4")

        warning_0 = "AngleCalculationModule has not been tested for SPHERE/IFS data."

        warning_1 = (
            "For SPHERE data it is recommended to use the header keyword 'ESO INS4 "
            "DROT2 RA' to specify the object's right ascension. The input will be "
            "parsed accordingly. Using the regular 'RA' keyword will lead to wrong "
            "parallactic angles."
        )

        warning_2 = (
            "For SPHERE data it is recommended to use the header keyword 'ESO INS4 "
            "DROT2 DEC' to specify the object's declination. The input will be parsed "
            "accordingly. Using the regular 'DEC' keyword will lead to wrong "
            "parallactic angles."
        )

        if len(warning) == 3:
            assert warning[0].message.args[0] == warning_0
            assert warning[1].message.args[0] == warning_1
            assert warning[2].message.args[0] == warning_2

        data = self.pipeline.get_data("header_read/PARANG")
        assert np.sum(data) == pytest.approx(
            -890.8506514377593, rel=self.limit, abs=0.0
        )
        assert data.shape == (10,)

    def test_angle_sort(self) -> None:

        index = self.pipeline.get_data("header_read/INDEX")
        self.pipeline.set_attribute("read", "INDEX", index[::-1], static=False)

        module = SortParangModule(
            name_in="sort1", image_in_tag="read", image_out_tag="read_sorted"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("sort1")
        self.pipeline.set_attribute("read", "INDEX", index, static=False)

        parang = self.pipeline.get_data("header_read/PARANG")[::-1]
        parang_sort = self.pipeline.get_data("header_read_sorted/PARANG")
        assert np.sum(parang) == pytest.approx(
            np.sum(parang_sort), rel=self.limit, abs=0.0
        )

        parang_set = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        self.pipeline.set_attribute("read_ifs", "PARANG", parang_set, static=False)

        data = self.pipeline.get_data("read_sorted")
        assert np.sum(data[0]) == pytest.approx(
            9.71156815235485, rel=self.limit, abs=0.0
        )

    def test_angle_sort_ifs(self) -> None:

        index = self.pipeline.get_data("header_read_ifs/INDEX")
        self.pipeline.set_attribute("read_ifs", "INDEX", index[::-1], static=False)

        module = SortParangModule(
            name_in="sort2", image_in_tag="read_ifs", image_out_tag="read_ifs_sorted"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("sort2")
        self.pipeline.set_attribute("read_ifs", "INDEX", index, static=False)

        parang = self.pipeline.get_data("header_read_ifs/PARANG")[::-1]
        parang_sort = self.pipeline.get_data("header_read_ifs_sorted/PARANG")
        assert np.sum(parang) == pytest.approx(
            np.sum(parang_sort), rel=self.limit, abs=0.0
        )

        data = self.pipeline.get_data("read_ifs_sorted")
        assert np.sum(data[0, 0]) == pytest.approx(
            21.185139976163477, rel=self.limit, abs=0.0
        )

    def test_angle_interpolation_mismatch(self) -> None:

        self.pipeline.set_attribute("read", "NDIT", [9, 9, 9, 9], static=False)

        module = AngleInterpolationModule(name_in="angle5", data_tag="read")

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("angle5")

        warning_0 = (
            "There is a mismatch between the NDIT and NFRAMES values. The parallactic "
            "angles are calculated with a linear interpolation by using NFRAMES steps. "
            "A frame selection should be applied after the parallactic angles are "
            "calculated."
        )

        if len(warning) == 1:
            assert warning[0].message.args[0] == warning_0

        data = self.pipeline.get_data("header_read/PARANG")
        assert np.sum(data) == pytest.approx(900.0, rel=self.limit, abs=0.0)
        assert data.shape == (10,)

    def test_psf_preparation_norm_mask(self) -> None:

        module = PSFpreparationModule(
            name_in="prep1",
            image_in_tag="read",
            image_out_tag="prep1",
            mask_out_tag="mask1",
            norm=True,
            cent_size=0.1,
            edge_size=1.0,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("prep1")

        data = self.pipeline.get_data("prep1")
        assert np.sum(data) == pytest.approx(
            -1.5844830188044685, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_data("mask1")
        assert np.sum(data) == pytest.approx(52, rel=self.limit, abs=0.0)
        assert data.shape == (11, 11)

    def test_psf_preparation_none(self) -> None:

        module = PSFpreparationModule(
            name_in="prep2",
            image_in_tag="read",
            image_out_tag="prep2",
            mask_out_tag="mask2",
            norm=False,
            cent_size=None,
            edge_size=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("prep2")

        data = self.pipeline.get_data("prep2")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_psf_preparation_no_mask_out(self) -> None:

        module = PSFpreparationModule(
            name_in="prep3",
            image_in_tag="read",
            image_out_tag="prep3",
            mask_out_tag=None,
            norm=False,
            cent_size=None,
            edge_size=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("prep3")

        data = self.pipeline.get_data("prep3")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_psf_preparation_sdi(self) -> None:

        module = PSFpreparationModule(
            name_in="prep4",
            image_in_tag="read_ifs",
            image_out_tag="prep4",
            mask_out_tag=None,
            norm=False,
            cent_size=None,
            edge_size=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("prep4")

        data = self.pipeline.get_data("prep4")
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.0)
        assert data.shape == (3, 10, 21, 21)

    def test_sdi_preparation(self) -> None:

        module = SDIpreparationModule(
            name_in="sdi",
            wavelength=(0.65, 0.6),
            width=(0.1, 0.5),
            image_in_tag="read",
            image_out_tag="sdi",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("sdi")

        data = self.pipeline.get_data("sdi")
        assert np.sum(data) == pytest.approx(
            21.084666133914183, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        attribute = self.pipeline.get_attribute("sdi", "History: SDIpreparationModule")
        assert attribute == "(line, continuum) = (0.65, 0.6)"



"""
Functions for combining the residuals of the PSF subtraction.
"""

from typing import Optional

import numpy as np

from typeguard import typechecked
from scipy.ndimage import rotate


@typechecked
def combine_residuals(
    method: str,
    res_rot: np.ndarray,
    residuals: Optional[np.ndarray] = None,
    angles: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Wavelength wrapper for the combine_residual function. Produces an array with either 1
    or number of wavelengths sized array.

    Parameters
    ----------
    method : str
        Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    res_rot : np.ndarray
        Derotated residuals of the PSF subtraction (3D).
    residuals : np.ndarray, None
        Non-derotated residuals of the PSF subtraction (3D). Only required for the noise-weighted
        residuals.
    angles : np.ndarray, None
        Derotation angles (deg). Only required for the noise-weighted residuals.

    Returns
    -------
    np.ndarray
        Collapsed residuals (3D).
    """

    if res_rot.ndim == 3:
        output = _residuals(
            method=method,
            res_rot=np.asarray(res_rot),
            residuals=residuals,
            angles=angles,
        )

    if res_rot.ndim == 4:
        output = np.zeros((res_rot.shape[0], res_rot.shape[2], res_rot.shape[3]))

        for i in range(res_rot.shape[0]):
            if residuals is None:
                output[i,] = _residuals(
                    method=method,
                    res_rot=res_rot[i,],
                    residuals=residuals,
                    angles=angles,
                )[0]

            else:
                output[i,] = _residuals(
                    method=method,
                    res_rot=res_rot[i,],
                    residuals=residuals[i,],
                    angles=angles,
                )[0]

    return output


@typechecked
def _residuals(
    method: str,
    res_rot: np.ndarray,
    residuals: Optional[np.ndarray] = None,
    angles: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Function for combining the derotated residuals of the PSF subtraction.

    Parameters
    ----------
    method : str
        Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    res_rot : np.ndarray
        Derotated residuals of the PSF subtraction (3D).
    residuals : np.ndarray, None
        Non-derotated residuals of the PSF subtraction (3D). Only required for the noise-weighted
        residuals.
    angles : np.ndarray, None
        Derotation angles (deg). Only required for the noise-weighted residuals.

    Returns
    -------
    np.ndarray
        Combined residuals (3D).
    """

    if method == "mean":
        stack = np.mean(res_rot, axis=0)

    elif method == "median":
        stack = np.median(res_rot, axis=0)

    elif method == "weighted":
        tmp_res_var = np.var(residuals, axis=0)

        res_repeat = np.repeat(
            tmp_res_var[np.newaxis, :, :], repeats=residuals.shape[0], axis=0
        )

        res_var = np.zeros(res_repeat.shape)

        for j, angle in enumerate(angles):
            # scipy.ndimage.rotate rotates in clockwise direction for positive angles
            res_var[j,] = rotate(input=res_repeat[j,], angle=angle, reshape=False)

        weight1 = np.divide(
            res_rot,
            res_var,
            out=np.zeros_like(res_var),
            where=(np.abs(res_var) > 1e-100) & (res_var != np.nan),
        )

        weight2 = np.divide(
            1.0,
            res_var,
            out=np.zeros_like(res_var),
            where=(np.abs(res_var) > 1e-100) & (res_var != np.nan),
        )

        sum1 = np.sum(weight1, axis=0)
        sum2 = np.sum(weight2, axis=0)

        stack = np.divide(
            sum1,
            sum2,
            out=np.zeros_like(sum2),
            where=(np.abs(sum2) > 1e-100) & (sum2 != np.nan),
        )

    elif method == "clipped":
        stack = np.zeros(res_rot.shape[-2:])

        for i in range(stack.shape[0]):
            for j in range(stack.shape[1]):
                pix_line = res_rot[:, i, j]

                if np.var(pix_line) > 0.0:
                    no_mean = pix_line - np.mean(pix_line)

                    part1 = no_mean.compress(
                        (no_mean < 3.0 * np.sqrt(np.var(no_mean))).flat
                    )
                    part2 = part1.compress(
                        (part1 > -3.0 * np.sqrt(np.var(no_mean))).flat
                    )

                    stack[i, j] = np.mean(pix_line) + np.mean(part2)

    return stack[np.newaxis, ...]



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.resizing import (
    CropImagesModule,
    ScaleImagesModule,
    AddLinesModule,
    RemoveLinesModule,
)
from pynpoint.util.tests import (
    create_config,
    create_star_data,
    create_ifs_data,
    remove_test_data,
)


class TestResizing:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "resize")
        create_ifs_data(self.test_dir + "resize_ifs")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["resize", "resize_ifs"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read",
            image_tag="read",
            input_dir=self.test_dir + "resize",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read")

        data = self.pipeline.get_data("read")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(
            name_in="read_ifs",
            image_tag="read_ifs",
            input_dir=self.test_dir + "resize_ifs",
            overwrite=True,
            check=True,
            ifs_data=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read_ifs")

        data = self.pipeline.get_data("read_ifs")
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.0)
        assert data.shape == (3, 10, 21, 21)

    def test_crop_images(self) -> None:

        module = CropImagesModule(
            size=0.2,
            center=None,
            name_in="crop1",
            image_in_tag="read",
            image_out_tag="crop1",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("crop1")

        module = CropImagesModule(
            size=0.2,
            center=(4, 4),
            name_in="crop2",
            image_in_tag="read",
            image_out_tag="crop2",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("crop2")

        module = CropImagesModule(
            size=0.2,
            center=(4, 4),
            name_in="crop_ifs",
            image_in_tag="read_ifs",
            image_out_tag="crop_ifs",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("crop_ifs")

        data = self.pipeline.get_data("crop1")
        assert np.sum(data) == pytest.approx(
            104.93318507061295, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

        data = self.pipeline.get_data("crop2")
        assert np.sum(data) == pytest.approx(
            105.64863165433025, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 9, 9)

        data = self.pipeline.get_data("crop_ifs")
        assert np.sum(data) == pytest.approx(
            15.870936600122521, rel=self.limit, abs=0.0
        )
        assert data.shape == (3, 10, 9, 9)

    def test_scale_images(self) -> None:

        module = ScaleImagesModule(
            name_in="scale1",
            image_in_tag="read",
            image_out_tag="scale1",
            scaling=(2.0, 2.0, None),
            pixscale=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("scale1")

        module = ScaleImagesModule(
            name_in="scale2",
            image_in_tag="read",
            image_out_tag="scale2",
            scaling=(None, None, 2.0),
            pixscale=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("scale2")

        data = self.pipeline.get_data("scale1")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 22, 22)

        data = self.pipeline.get_data("scale2")
        assert np.sum(data) == pytest.approx(
            211.08557759610554, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        attr = self.pipeline.get_attribute("read", "PIXSCALE", static=True)
        assert attr == pytest.approx(0.027, rel=self.limit, abs=0.0)

        attr = self.pipeline.get_attribute("scale1", "PIXSCALE", static=True)
        assert attr == pytest.approx(0.0135, rel=self.limit, abs=0.0)

        attr = self.pipeline.get_attribute("scale2", "PIXSCALE", static=True)
        assert attr == pytest.approx(0.027, rel=self.limit, abs=0.0)

    def test_add_lines(self) -> None:

        module = AddLinesModule(
            lines=(2, 5, 0, 3), name_in="add", image_in_tag="read", image_out_tag="add"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("add")

        data = self.pipeline.get_data("add")
        assert np.sum(data) == pytest.approx(
            105.54278879805275, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 14, 18)

    def test_remove_lines(self) -> None:

        module = RemoveLinesModule(
            lines=(2, 5, 0, 3),
            name_in="remove",
            image_in_tag="read",
            image_out_tag="remove",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("remove")

        data = self.pipeline.get_data("remove")
        assert np.sum(data) == pytest.approx(67.49726677462391, rel=self.limit, abs=0.0)
        assert data.shape == (10, 8, 4)

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 4

        self.pipeline.run_module("remove")

        data_multi = self.pipeline.get_data("remove")
        assert data == pytest.approx(data_multi, rel=self.limit, abs=0.0)
        assert data.shape == data_multi.shape



import os
import h5py

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.psfsubtraction import PcaPsfSubtractionModule
from pynpoint.util.tests import create_config, create_ifs_data, remove_test_data


class TestPsfSubtractionSdi:

    def setup_class(self) -> None:

        self.limit = 1e-5
        self.test_dir = os.path.dirname(__file__) + "/"

        create_ifs_data(self.test_dir + "science")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["science"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read",
            image_tag="science",
            input_dir=self.test_dir + "science",
            ifs_data=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read")

        data = self.pipeline.get_data("science")
        assert np.sum(data) == pytest.approx(749.8396528807368, rel=self.limit, abs=0.0)
        assert data.shape == (3, 10, 21, 21)

        self.pipeline.set_attribute(
            "science", "WAVELENGTH", [1.0, 1.1, 1.2], static=False
        )
        self.pipeline.set_attribute(
            "science", "PARANG", np.linspace(0.0, 180.0, 10), static=False
        )

    def test_psf_subtraction_sdi(self) -> None:

        processing_types = ["ADI", "SDI+ADI", "ADI+SDI", "CODI"]

        expected = [
            [
                -0.16718942968552664,
                -0.790697125718532,
                19.507979777136892,
                -0.21617058715490922,
            ],
            [
                -0.001347198747121658,
                -0.08621264803633322,
                2.3073192270025333,
                -0.010269745733878437,
            ],
            [
                0.009450917836998779,
                -0.05776205365084376,
                -0.43506678222476264,
                0.0058856438951644455,
            ],
            [-0.2428739554898396, -0.5069023645693083, 9.326414176548905, 0.00],
        ]

        shape_expc = [
            (2, 3, 21, 21),
            (2, 2, 3, 21, 21),
            (1, 1, 3, 21, 21),
            (2, 3, 21, 21),
        ]

        pca_numbers = [range(1, 3), (range(1, 3), range(1, 3)), ([1], [1]), range(1, 3)]

        res_arr_tags = [None, None, "res_arr_single_sdi_ADI+SDI", None]

        for i, p_type in enumerate(processing_types):

            module = PcaPsfSubtractionModule(
                pca_numbers=pca_numbers[i],
                name_in="pca_single_sdi_" + p_type,
                images_in_tag="science",
                reference_in_tag="science",
                res_mean_tag="res_mean_single_sdi_" + p_type,
                res_median_tag="res_median_single_sdi_" + p_type,
                res_weighted_tag="res_weighted_single_sdi_" + p_type,
                res_rot_mean_clip_tag="res_clip_single_sdi_" + p_type,
                res_arr_out_tag=res_arr_tags[i],
                basis_out_tag="basis_single_sdi_" + p_type,
                extra_rot=0.0,
                subtract_mean=True,
                processing_type=p_type,
            )

            self.pipeline.add_module(module)
            self.pipeline.run_module("pca_single_sdi_" + p_type)

            data = self.pipeline.get_data("res_mean_single_sdi_" + p_type)
            assert np.sum(data) == pytest.approx(
                expected[i][0], rel=self.limit, abs=0.0
            )
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data("res_median_single_sdi_" + p_type)
            assert np.sum(data) == pytest.approx(
                expected[i][1], rel=self.limit, abs=0.0
            )
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data("res_weighted_single_sdi_" + p_type)
            assert np.sum(data) == pytest.approx(
                expected[i][2], rel=self.limit, abs=0.0
            )
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data("res_clip_single_sdi_" + p_type)
            #            assert np.sum(data) == pytest.approx(expected[i][3], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            # data = self.pipeline.get_data('basis_single_sdi_'+p_type)
            # assert np.sum(data) == pytest.approx(-1.3886119555248766, rel=self.limit, abs=0.)
            # assert data.shape == (5, 30, 30)

    def test_psf_subtraction_sdi_multi(self) -> None:

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 4

        processing_types = ["SDI", "ADI+SDI"]

        pca_numbers = [range(1, 3), (range(1, 3), range(1, 3))]

        expected = [
            [
                -0.004159475403024583,
                0.02613693149969979,
                -0.12940723035023394,
                -0.008432530081399985,
            ],
            [
                -0.006580571531064533,
                -0.08171546066331437,
                0.5700432018961117,
                -0.014527353460544753,
            ],
        ]

        shape_expc = [(2, 3, 21, 21), (2, 2, 3, 21, 21)]

        for i, p_type in enumerate(processing_types):

            module = PcaPsfSubtractionModule(
                pca_numbers=pca_numbers[i],
                name_in="pca_multi_sdi_" + p_type,
                images_in_tag="science",
                reference_in_tag="science",
                res_mean_tag="res_mean_multi_sdi_" + p_type,
                res_median_tag="res_median_multi_sdi_" + p_type,
                res_weighted_tag="res_weighted_multi_sdi_" + p_type,
                res_rot_mean_clip_tag="res_clip_multi_sdi_" + p_type,
                res_arr_out_tag=None,
                basis_out_tag=None,
                extra_rot=0.0,
                subtract_mean=True,
                processing_type=p_type,
            )

            self.pipeline.add_module(module)
            self.pipeline.run_module("pca_multi_sdi_" + p_type)

            data = self.pipeline.get_data("res_mean_multi_sdi_" + p_type)
            assert np.sum(data) == pytest.approx(
                expected[i][0], rel=self.limit, abs=0.0
            )
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data("res_median_multi_sdi_" + p_type)
            assert np.sum(data) == pytest.approx(
                expected[i][1], rel=self.limit, abs=0.0
            )
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data("res_weighted_multi_sdi_" + p_type)
            assert np.sum(data) == pytest.approx(
                expected[i][2], rel=self.limit, abs=0.0
            )
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data("res_clip_multi_sdi_" + p_type)
            #            assert np.sum(data) == pytest.approx(expected[i][3], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.textwriting import TextWritingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestTextWriting:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_random(self.test_dir, nimages=1)
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=["image.dat", "data.dat"])

    def test_input_data(self) -> None:

        data = self.pipeline.get_data("images")
        assert np.sum(data) == pytest.approx(
            0.0008557279524431129, rel=self.limit, abs=0.0
        )
        assert data.shape == (1, 11, 11)

    def test_text_writing(self) -> None:

        text_write = TextWritingModule(
            name_in="text_write",
            data_tag="images",
            file_name="image.dat",
            output_dir=None,
            header=None,
        )

        self.pipeline.add_module(text_write)
        self.pipeline.run_module("text_write")

        data = np.loadtxt(self.test_dir + "image.dat")

        assert np.sum(data) == pytest.approx(
            0.0008557279524431129, rel=self.limit, abs=0.0
        )
        assert data.shape == (11, 11)

    def test_text_writing_ndim(self) -> None:

        data_4d = np.random.normal(loc=0, scale=2e-4, size=(5, 5, 5, 5))

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file.create_dataset("data_4d", data=data_4d)

        text_write = TextWritingModule(
            name_in="write_4d",
            data_tag="data_4d",
            file_name="data.dat",
            output_dir=None,
            header=None,
        )

        self.pipeline.add_module(text_write)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("write_4d")

        assert str(error.value) == "Only 1D or 2D arrays can be written to a text file."

    def test_text_writing_int(self) -> None:

        data_int = np.arange(1, 11, 1)

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file.create_dataset("data_int", data=data_int)

        text_write = TextWritingModule(
            name_in="write_int",
            data_tag="data_int",
            file_name="data.dat",
            output_dir=None,
            header=None,
        )

        self.pipeline.add_module(text_write)
        self.pipeline.run_module("write_int")

        data = np.loadtxt(self.test_dir + "data.dat")

        assert data == pytest.approx(data_int, rel=self.limit, abs=0.0)
        assert data.shape == (10,)



import os
import math

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.processing.badpixel import (
    BadPixelSigmaFilterModule,
    BadPixelMapModule,
    BadPixelInterpolationModule,
    BadPixelTimeFilterModule,
    ReplaceBadPixelsModule,
)
from pynpoint.util.tests import create_config, remove_test_data


class TestBadPixel:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        np.random.seed(1)

        images = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        dark = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        flat = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))

        images[0, 5, 5] = 1.0
        dark[:, 5, 5] = 1.0
        flat[:, 8, 8] = -1.0
        flat[:, 9, 9] = -1.0
        flat[:, 10, 10] = -1.0

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "w") as hdf_file:
            hdf_file.create_dataset("images", data=images)
            hdf_file.create_dataset("dark", data=dark)
            hdf_file.create_dataset("flat", data=flat)

        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_bad_pixel_sigma_filter(self) -> None:

        module = BadPixelSigmaFilterModule(
            name_in="sigma1",
            image_in_tag="images",
            image_out_tag="sigma1",
            map_out_tag="None",
            box=9,
            sigma=3.0,
            iterate=5,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("sigma1")

        data = self.pipeline.get_data("sigma1")
        assert np.sum(data) == pytest.approx(
            0.006513475520308432, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

    def test_bad_pixel_map_out(self) -> None:

        module = BadPixelSigmaFilterModule(
            name_in="sigma2",
            image_in_tag="images",
            image_out_tag="sigma2",
            map_out_tag="bpmap",
            box=9,
            sigma=2.0,
            iterate=3,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("sigma2")

        data = self.pipeline.get_data("sigma2")
        assert data[0, 0, 0] == pytest.approx(
            -2.4570591355257687e-05, rel=self.limit, abs=0.0
        )
        assert data[0, 5, 5] == pytest.approx(
            9.903775276151606e-06, rel=self.limit, abs=0.0
        )
        assert np.sum(data) == pytest.approx(
            0.011777887008566097, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

        data = self.pipeline.get_data("bpmap")
        assert data[0, 1, 1] == pytest.approx(1.0, rel=self.limit, abs=0.0)
        assert data[0, 5, 5] == pytest.approx(0.0, rel=self.limit, abs=0.0)
        assert np.sum(data) == pytest.approx(519.0, rel=self.limit, abs=0.0)
        assert data.shape == (5, 11, 11)

    def test_bad_pixel_map(self) -> None:

        module = BadPixelMapModule(
            name_in="bp_map1",
            dark_in_tag="dark",
            flat_in_tag="flat",
            bp_map_out_tag="bp_map1",
            dark_threshold=0.99,
            flat_threshold=-0.99,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("bp_map1")

        data = self.pipeline.get_data("bp_map1")
        assert data[0, 0, 0] == pytest.approx(1.0, rel=self.limit, abs=0.0)
        assert data[0, 5, 5] == pytest.approx(0.0, rel=self.limit, abs=0.0)
        assert np.sum(data) == pytest.approx(117.0, rel=self.limit, abs=0.0)
        assert data.shape == (1, 11, 11)

    def test_map_no_dark(self) -> None:

        module = BadPixelMapModule(
            name_in="bp_map2",
            dark_in_tag=None,
            flat_in_tag="flat",
            bp_map_out_tag="bp_map2",
            dark_threshold=0.99,
            flat_threshold=-0.99,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("bp_map2")

        data = self.pipeline.get_data("bp_map2")
        assert np.sum(data) == pytest.approx(118.0, rel=self.limit, abs=0.0)
        assert data.shape == (1, 11, 11)

    def test_map_no_flat(self) -> None:

        module = BadPixelMapModule(
            name_in="bp_map3",
            dark_in_tag="dark",
            flat_in_tag=None,
            bp_map_out_tag="bp_map3",
            dark_threshold=0.99,
            flat_threshold=-0.99,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("bp_map3")

        data = self.pipeline.get_data("bp_map3")
        assert np.sum(data) == pytest.approx(120.0, rel=self.limit, abs=0.0)
        assert data.shape == (1, 11, 11)

    def test_bad_pixel_interpolation(self) -> None:

        module = BadPixelInterpolationModule(
            name_in="interpolation",
            image_in_tag="images",
            bad_pixel_map_tag="bp_map1",
            image_out_tag="interpolation",
            iterations=10,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("interpolation")

        data = self.pipeline.get_data("interpolation")
        assert data[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=1e-6, abs=0.0)
        assert data[0, 5, 5] == pytest.approx(
            -1.4292408645473845e-05, rel=1e-6, abs=0.0
        )
        assert np.sum(data) == pytest.approx(0.008683344127872174, rel=1e-6, abs=0.0)
        assert data.shape == (5, 11, 11)

    def test_bad_pixel_time(self) -> None:

        module = BadPixelTimeFilterModule(
            name_in="time",
            image_in_tag="images",
            image_out_tag="time",
            sigma=(2.0, 2.0),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("time")

        data = self.pipeline.get_data("time")
        assert data[0, 0, 0] == pytest.approx(
            0.00032486907273264834, rel=self.limit, abs=0.0
        )
        assert data[0, 5, 5] == pytest.approx(
            -0.00017468532119886812, rel=self.limit, abs=0.0
        )
        assert np.sum(data) == pytest.approx(
            0.004175672043832705, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

    def test_replace_bad_pixels(self) -> None:

        module = ReplaceBadPixelsModule(
            name_in="replace1",
            image_in_tag="images",
            map_in_tag="bp_map1",
            image_out_tag="replace",
            size=2,
            replace="mean",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("replace1")

        data = self.pipeline.get_data("replace")
        assert data[0, 0, 0] == pytest.approx(
            0.00032486907273264834, rel=self.limit, abs=0.0
        )
        assert data[0, 5, 5] == pytest.approx(
            4.260114004413933e-05, rel=self.limit, abs=0.0
        )
        assert np.sum(data) == pytest.approx(
            0.00883357395370896, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

        module = ReplaceBadPixelsModule(
            name_in="replace2",
            image_in_tag="images",
            map_in_tag="bp_map1",
            image_out_tag="replace",
            size=2,
            replace="median",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("replace2")

        data = self.pipeline.get_data("replace")
        assert data[0, 0, 0] == pytest.approx(
            0.00032486907273264834, rel=self.limit, abs=0.0
        )
        assert data[0, 5, 5] == pytest.approx(
            4.327154179438619e-05, rel=self.limit, abs=0.0
        )
        assert np.sum(data) == pytest.approx(
            0.008489525337709688, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

        module = ReplaceBadPixelsModule(
            name_in="replace3",
            image_in_tag="images",
            map_in_tag="bp_map1",
            image_out_tag="replace",
            size=2,
            replace="nan",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("replace3")

        data = self.pipeline.get_data("replace")
        assert data[0, 0, 0] == pytest.approx(
            0.00032486907273264834, rel=self.limit, abs=0.0
        )
        assert math.isnan(data[0, 5, 5])
        assert np.nansum(data) == pytest.approx(
            0.009049653234723834, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.psfpreparation import PSFpreparationModule
from pynpoint.processing.psfsubtraction import (
    PcaPsfSubtractionModule,
    ClassicalADIModule,
)
from pynpoint.util.tests import create_config, create_fake_data, remove_test_data


class TestPsfSubtractionAdi:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_fake_data(self.test_dir + "science")
        create_fake_data(self.test_dir + "reference")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["science", "reference"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read1", image_tag="science", input_dir=self.test_dir + "science"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read1")

        data = self.pipeline.get_data("science")
        assert np.sum(data) == pytest.approx(
            11.012854046962481, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 21, 21)

        self.pipeline.set_attribute(
            "science", "PARANG", np.linspace(0.0, 180.0, 10), static=False
        )

        module = FitsReadingModule(
            name_in="read2",
            image_tag="reference",
            input_dir=self.test_dir + "reference",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read2")

        data = self.pipeline.get_data("reference")
        assert np.sum(data) == pytest.approx(
            11.012854046962481, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 21, 21)

    def test_psf_preparation(self) -> None:

        module = PSFpreparationModule(
            name_in="prep1",
            image_in_tag="science",
            image_out_tag="science_prep",
            mask_out_tag=None,
            norm=False,
            resize=None,
            cent_size=0.05,
            edge_size=1.0,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("prep1")

        data = self.pipeline.get_data("science_prep")
        assert np.sum(data) == pytest.approx(5.029285028467547, rel=self.limit, abs=0.0)
        assert data.shape == (10, 21, 21)

        module = PSFpreparationModule(
            name_in="prep2",
            image_in_tag="reference",
            image_out_tag="reference_prep",
            mask_out_tag=None,
            norm=False,
            resize=None,
            cent_size=0.05,
            edge_size=1.0,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("prep2")

        data = self.pipeline.get_data("reference_prep")
        assert np.sum(data) == pytest.approx(5.029285028467547, rel=self.limit, abs=0.0)
        assert data.shape == (10, 21, 21)

    def test_classical_adi(self) -> None:

        module = ClassicalADIModule(
            threshold=None,
            nreference=None,
            residuals="mean",
            extra_rot=0.0,
            name_in="cadi1",
            image_in_tag="science_prep",
            res_out_tag="cadi_res",
            stack_out_tag="cadi_stack",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("cadi1")

        data = self.pipeline.get_data("cadi_res")
        assert np.sum(data) == pytest.approx(
            0.8381625719865213, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data("cadi_stack")
        assert np.sum(data) == pytest.approx(
            0.08395606034388256, rel=self.limit, abs=0.0
        )
        assert data.shape == (1, 21, 21)

    def test_classical_adi_threshold(self) -> None:

        module = ClassicalADIModule(
            threshold=(0.1, 0.03, 1.0),
            nreference=5,
            residuals="median",
            extra_rot=0.0,
            name_in="cadi2",
            image_in_tag="science_prep",
            res_out_tag="cadi_res",
            stack_out_tag="cadi_stack",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("cadi2")

        data = self.pipeline.get_data("cadi_res")
        assert np.sum(data) == pytest.approx(0.7158207863548083, rel=1e-2, abs=0.0)
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data("cadi_stack")
        assert np.sum(data) == pytest.approx(0.07448334552227256, rel=1e-2, abs=0.0)
        assert data.shape == (1, 21, 21)

    def test_psf_subtraction_pca_single(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_single",
            images_in_tag="science",
            reference_in_tag="science",
            res_mean_tag="res_mean_single",
            res_median_tag="res_median_single",
            res_weighted_tag="res_weighted_single",
            res_rot_mean_clip_tag="res_clip_single",
            res_arr_out_tag="res_arr_single",
            basis_out_tag="basis_single",
            extra_rot=45.0,
            subtract_mean=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_single")

        data = self.pipeline.get_data("res_mean_single")
        assert np.sum(data) == pytest.approx(
            -0.00011857022709778602, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_median_single")
        assert np.sum(data) == pytest.approx(
            -0.002184868916566093, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_weighted_single")
        assert np.sum(data) == pytest.approx(
            0.08102176735226937, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        # data = self.pipeline.get_data('res_clip_single')
        # assert np.sum(data) == pytest.approx(7.09495495339349e-05, rel=self.limit, abs=0.)
        # assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_arr_single1")
        assert np.sum(data) == pytest.approx(
            -0.0002751385418691618, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data("basis_single")
        assert np.sum(data) == pytest.approx(
            0.09438697731322143, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_no_mean(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_no_mean",
            images_in_tag="science",
            reference_in_tag="science",
            res_mean_tag="res_mean_no_mean",
            res_median_tag=None,
            res_weighted_tag=None,
            res_rot_mean_clip_tag=None,
            res_arr_out_tag=None,
            basis_out_tag="basis_no_mean",
            extra_rot=0.0,
            subtract_mean=False,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_no_mean")

        data = self.pipeline.get_data("res_mean_no_mean")
        assert np.sum(data) == pytest.approx(
            0.0006081272007585688, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("basis_no_mean")
        assert np.sum(data) == pytest.approx(5.118005177367776, rel=self.limit, abs=0.0)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_ref",
            images_in_tag="science",
            reference_in_tag="reference",
            res_mean_tag="res_mean_ref",
            res_median_tag=None,
            res_weighted_tag=None,
            res_rot_mean_clip_tag=None,
            res_arr_out_tag=None,
            basis_out_tag="basis_ref",
            extra_rot=0.0,
            subtract_mean=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_ref")

        data = self.pipeline.get_data("res_mean_ref")
        assert np.sum(data) == pytest.approx(
            0.0006330226118859073, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("basis_ref")
        assert np.sum(data) == pytest.approx(
            0.0943869773132221, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref_no_mean(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_ref_no_mean",
            images_in_tag="science",
            reference_in_tag="reference",
            res_mean_tag="res_mean_ref_no_mean",
            res_median_tag=None,
            res_weighted_tag=None,
            res_rot_mean_clip_tag=None,
            res_arr_out_tag=None,
            basis_out_tag="basis_ref_no_mean",
            extra_rot=0.0,
            subtract_mean=False,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_ref_no_mean")

        data = self.pipeline.get_data("res_mean_ref_no_mean")
        assert np.sum(data) == pytest.approx(
            0.0006081272007585764, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("basis_ref_no_mean")
        assert np.sum(data) == pytest.approx(5.118005177367774, rel=self.limit, abs=0.0)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_pca_single_mask(self) -> None:

        pca = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_single_mask",
            images_in_tag="science_prep",
            reference_in_tag="science_prep",
            res_mean_tag="res_mean_single_mask",
            res_median_tag="res_median_single_mask",
            res_weighted_tag="res_weighted_single_mask",
            res_rot_mean_clip_tag="res_clip_single_mask",
            res_arr_out_tag="res_arr_single_mask",
            basis_out_tag="basis_single_mask",
            extra_rot=45.0,
            subtract_mean=True,
        )

        self.pipeline.add_module(pca)
        self.pipeline.run_module("pca_single_mask")

        data = self.pipeline.get_data("res_mean_single_mask")
        assert np.sum(data) == pytest.approx(
            0.00010696166038626307, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_median_single_mask")
        assert np.sum(data) == pytest.approx(
            -0.0021005307611346156, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_weighted_single_mask")
        assert np.sum(data) == pytest.approx(
            0.06014309988789256, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_clip_single_mask")
        # assert np.sum(data) == pytest.approx(9.35120662148806e-05, rel=self.limit, abs=0.)
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("res_arr_single_mask1")
        assert np.sum(data) == pytest.approx(
            0.0006170872862547557, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 21, 21)

        data = self.pipeline.get_data("basis_single_mask")
        assert np.sum(data) == pytest.approx(
            0.08411251293842359, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_no_mean_mask(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_no_mean_mask",
            images_in_tag="science_prep",
            reference_in_tag="science_prep",
            res_mean_tag="res_mean_no_mean_mask",
            res_median_tag=None,
            res_weighted_tag=None,
            res_rot_mean_clip_tag=None,
            res_arr_out_tag=None,
            basis_out_tag="basis_no_mean_mask",
            extra_rot=0.0,
            subtract_mean=False,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_no_mean_mask")

        data = self.pipeline.get_data("res_mean_no_mean_mask")
        assert np.sum(data) == pytest.approx(
            2.3542359949502915e-05, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("basis_no_mean_mask")
        assert np.sum(data) == pytest.approx(5.655460951633232, rel=self.limit, abs=0.0)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref_mask(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_ref_mask",
            images_in_tag="science_prep",
            reference_in_tag="reference_prep",
            res_mean_tag="res_mean_ref_mask",
            res_median_tag=None,
            res_weighted_tag=None,
            res_rot_mean_clip_tag=None,
            res_arr_out_tag=None,
            basis_out_tag="basis_ref_mask",
            extra_rot=0.0,
            subtract_mean=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_ref_mask")

        data = self.pipeline.get_data("res_mean_ref_mask")
        assert np.sum(data) == pytest.approx(
            9.400558926815758e-06, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("basis_ref_mask")
        assert np.sum(data) == pytest.approx(
            0.08411251293842326, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_ref_no_mean_mask(self) -> None:

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_ref_no_mean_mask",
            images_in_tag="science_prep",
            reference_in_tag="reference_prep",
            res_mean_tag="res_mean_ref_no_mean_mask",
            res_median_tag=None,
            res_weighted_tag=None,
            res_rot_mean_clip_tag=None,
            res_arr_out_tag=None,
            basis_out_tag="basis_ref_no_mean_mask",
            extra_rot=0.0,
            subtract_mean=False,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_ref_no_mean_mask")

        data = self.pipeline.get_data("res_mean_ref_no_mean_mask")
        assert np.sum(data) == pytest.approx(
            2.354235994950671e-05, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 21, 21)

        data = self.pipeline.get_data("basis_ref_no_mean_mask")
        assert np.sum(data) == pytest.approx(5.655460951633233, rel=self.limit, abs=0.0)
        assert data.shape == (2, 21, 21)

    def test_psf_subtraction_pca_multi(self) -> None:

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 4

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_multi",
            images_in_tag="science",
            reference_in_tag="science",
            res_mean_tag="res_mean_multi",
            res_median_tag="res_median_multi",
            res_weighted_tag="res_weighted_multi",
            res_rot_mean_clip_tag="res_clip_multi",
            res_arr_out_tag=None,
            basis_out_tag="basis_multi",
            extra_rot=45.0,
            subtract_mean=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_multi")

        data_single = self.pipeline.get_data("res_mean_single")
        data_multi = self.pipeline.get_data("res_mean_multi")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

        data_single = self.pipeline.get_data("res_median_single")
        data_multi = self.pipeline.get_data("res_median_multi")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

        data_single = self.pipeline.get_data("res_weighted_single")
        data_multi = self.pipeline.get_data("res_weighted_multi")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

        data_single = self.pipeline.get_data("basis_single")
        data_multi = self.pipeline.get_data("basis_multi")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

    def test_psf_subtraction_pca_multi_mask(self) -> None:

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 4

        module = PcaPsfSubtractionModule(
            pca_numbers=range(1, 3),
            name_in="pca_multi_mask",
            images_in_tag="science_prep",
            reference_in_tag="science_prep",
            res_mean_tag="res_mean_multi_mask",
            res_median_tag="res_median_multi_mask",
            res_weighted_tag="res_weighted_multi_mask",
            res_rot_mean_clip_tag="res_clip_multi_mask",
            res_arr_out_tag=None,
            basis_out_tag="basis_multi_mask",
            extra_rot=45.0,
            subtract_mean=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("pca_multi_mask")

        data_single = self.pipeline.get_data("res_mean_single_mask")
        data_multi = self.pipeline.get_data("res_mean_multi_mask")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

        data_single = self.pipeline.get_data("res_median_single_mask")
        data_multi = self.pipeline.get_data("res_median_multi_mask")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

        data_single = self.pipeline.get_data("res_weighted_single_mask")
        data_multi = self.pipeline.get_data("res_weighted_multi_mask")
        assert data_single.shape == data_multi.shape

        assert data_single[data_single > 1e-12] == pytest.approx(
            data_multi[data_multi > 1e-12], rel=self.limit, abs=0.0
        )

        data_single = self.pipeline.get_data("basis_single_mask")
        data_multi = self.pipeline.get_data("basis_multi_mask")
        assert data_single.shape == data_multi.shape
        assert data_single == pytest.approx(data_multi, rel=self.limit, abs=0.0)

    def test_psf_subtraction_len_parang(self) -> None:

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file["config"].attrs["CPU"] = 1

        parang = self.pipeline.get_data("header_science/PARANG")
        self.pipeline.set_attribute(
            "science_prep", "PARANG", np.append(parang, 0.0), static=False
        )

        module = PcaPsfSubtractionModule(
            pca_numbers=[
                1,
            ],
            name_in="pca_len_parang",
            images_in_tag="science_prep",
            reference_in_tag="science_prep",
            res_mean_tag="res_mean_len_parang",
            extra_rot=0.0,
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("pca_len_parang")

        assert (
            str(error.value)
            == "The number of images (10) is not equal to the number of "
            "parallactic angles (11)."
        )



import os

import pytest

from pynpoint.core.pypeline import Pypeline
from pynpoint.core.dataio import ConfigPort, DataStorage
from pynpoint.util.tests import create_config, remove_test_data


class TestConfigPort:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"
        create_config(self.test_dir + "PynPoint_config.ini")

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_create_config_port(self) -> None:

        storage = DataStorage(self.test_dir + "PynPoint_database.hdf5")

        with pytest.raises(ValueError) as error:
            ConfigPort("images", storage)

        assert (
            str(error.value) == "The tag name of the central configuration should be "
            "'config'."
        )

        port = ConfigPort("config", None)

        with pytest.warns(UserWarning) as warning:
            check_error = port._check_error_cases()

        assert len(warning) == 1

        assert (
            warning[0].message.args[0]
            == "ConfigPort can not load data unless a database is "
            "connected."
        )

        assert not check_error

        port = ConfigPort("config", storage)
        assert isinstance(port, ConfigPort)

        with pytest.warns(UserWarning) as warning:
            port._check_error_cases()

        assert len(warning) == 1

        assert (
            warning[0].message.args[0]
            == "No data under the tag which is linked by the "
            "ConfigPort."
        )

    def test_get_config_attribute(self) -> None:

        create_config(self.test_dir + "PynPoint_config.ini")
        Pypeline(self.test_dir, self.test_dir, self.test_dir)

        storage = DataStorage(self.test_dir + "PynPoint_database.hdf5")
        port = ConfigPort("config", None)

        with pytest.warns(UserWarning) as warning:
            attribute = port.get_attribute("CPU")

        assert len(warning) == 1

        assert (
            warning[0].message.args[0]
            == "ConfigPort can not load data unless a database is "
            "connected."
        )

        assert attribute is None

        port = ConfigPort("config", storage)

        attribute = port.get_attribute("CPU")
        assert attribute == 1

        attribute = port.get_attribute("NFRAMES")
        assert attribute == "NAXIS3"

        attribute = port.get_attribute("PIXSCALE")
        assert attribute == pytest.approx(0.027, rel=self.limit, abs=0.0)

        with pytest.warns(UserWarning) as warning:
            attribute = port.get_attribute("test")

        assert len(warning) == 1

        assert warning[0].message.args[0] == "The attribute 'test' was not found."

        assert attribute is None



import os

import pytest
import numpy as np

from pynpoint.core.dataio import DataStorage, InputPort, OutputPort
from pynpoint.util.tests import create_random, create_config, remove_test_data


class TestInputPort:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_random(self.test_dir)
        create_config(self.test_dir + "PynPoint_config.ini")

        file_in = os.path.dirname(__file__) + "/PynPoint_database.hdf5"

        self.storage = DataStorage(file_in)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_create_instance_access_data(self) -> None:

        with pytest.raises(ValueError) as error:
            InputPort("config", self.storage)

        assert (
            str(error.value) == "The tag name 'config' is reserved for the central "
            "configuration of PynPoint."
        )

        with pytest.raises(ValueError) as error:
            InputPort("fits_header", self.storage)

        assert (
            str(error.value)
            == "The tag name 'fits_header' is reserved for storage of the "
            "FITS headers."
        )

        port = InputPort("images", self.storage)

        assert port[0, 0, 0] == pytest.approx(
            0.00032486907273264834, rel=self.limit, abs=0.0
        )

        data = np.mean(port.get_all())
        assert data == pytest.approx(1.1824138000882435e-05, rel=self.limit, abs=0.0)

        assert len(port[0:2, 0, 0]) == 2
        assert port.get_shape() == (5, 11, 11)

        assert port.get_attribute("PIXSCALE") == 0.01
        assert port.get_attribute("PARANG")[0] == 0.0

        with pytest.warns(UserWarning):
            assert port.get_attribute("none") is None

    def test_create_instance_access_non_existing_data(self) -> None:

        port = InputPort("test", self.storage)

        with pytest.warns(UserWarning):
            assert port[0, 0, 0] is None

        with pytest.warns(UserWarning):
            assert port.get_all() is None

        with pytest.warns(UserWarning):
            assert port.get_shape() is None

        with pytest.warns(UserWarning):
            assert port.get_attribute("num_files") is None

        with pytest.warns(UserWarning):
            assert port.get_all_non_static_attributes() is None

        with pytest.warns(UserWarning):
            assert port.get_all_static_attributes() is None

    def test_create_instance_no_data_storage(self) -> None:

        port = InputPort("test")

        with pytest.warns(UserWarning):
            assert port[0, 0, 0] is None

        with pytest.warns(UserWarning):
            assert port.get_all() is None

        with pytest.warns(UserWarning):
            assert port.get_shape() is None

        with pytest.warns(UserWarning):
            assert port.get_all_non_static_attributes() is None

        with pytest.warns(UserWarning):
            assert port.get_all_static_attributes() is None

    # def test_get_all_attributes(self) -> None:
    #
    #     port = InputPort('images', self.storage)
    #
    #     assert port.get_all_static_attributes() == {'PIXSCALE': 0.01}
    #     assert port.get_all_non_static_attributes() == ['PARANG', ]
    #
    #     port = OutputPort('images', self.storage)
    #     assert port.del_all_attributes() is None
    #
    #     port = InputPort('images', self.storage)
    #     assert port.get_all_non_static_attributes() is None

    # def test_get_ndim(self) -> None:
    #
    #     with pytest.warns(UserWarning) as warning:
    #         ndim = InputPort('images', None).get_ndim()
    #
    #     assert len(warning) == 1
    #
    #     assert warning[0].message.args[0] == 'InputPort can not load data ' \
    #                                          'unless a database is connected.'
    #
    #     assert ndim is None
    #
    #     port = InputPort('images', self.storage)
    #     assert port.get_ndim() == 3



"""
Pipeline modules for basic image operations.
"""

import time

from typing import Tuple

from typeguard import typechecked
from scipy.ndimage import rotate

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.module import progress, memory_frames


class SubtractImagesModule(ProcessingModule):
    """
    Pipeline module for subtracting two sets of images.
    """

    __author__ = "Tomas Stolker"

    @typechecked
    def __init__(
        self,
        name_in: str,
        image_in_tags: Tuple[str, str],
        image_out_tag: str,
        scaling: float = 1.0,
    ) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tags : tuple(str, str)
            Tuple with two tags of the database entry that are read as input.
        image_out_tag : str
            Tag of the database entry with the subtracted images that are written as output.
        scaling : float
            Additional scaling factor.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in1_port = self.add_input_port(image_in_tags[0])
        self.m_image_in2_port = self.add_input_port(image_in_tags[1])
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_scaling = scaling

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Subtracts the images from the second database tag from the images
        of the first database tag, on a frame-by-frame basis.

        Returns
        -------
        NoneType
            None
        """

        if self.m_image_in1_port.get_shape() != self.m_image_in2_port.get_shape():
            raise ValueError("The shape of the two input tags has to be the same.")

        memory = self._m_config_port.get_attribute("MEMORY")
        nimages = self.m_image_in1_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), "Subtracting images...", start_time)

            images1 = self.m_image_in1_port[frames[i] : frames[i + 1],]
            images2 = self.m_image_in2_port[frames[i] : frames[i + 1],]

            self.m_image_out_port.append(
                self.m_scaling * (images1 - images2), data_dim=3
            )

        history = f"scaling = {self.m_scaling}"
        self.m_image_out_port.add_history("SubtractImagesModule", history)
        self.m_image_out_port.copy_attributes(self.m_image_in1_port)
        self.m_image_out_port.close_port()


class AddImagesModule(ProcessingModule):
    """
    Pipeline module for adding two sets of images.
    """

    __author__ = "Tomas Stolker"

    @typechecked
    def __init__(
        self,
        name_in: str,
        image_in_tags: Tuple[str, str],
        image_out_tag: str,
        scaling: float = 1.0,
    ) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tags : tuple(str, str)
            Tuple with two tags of the database entry that are read as input.
        image_out_tag : str
            Tag of the database entry with the added images that are written as output.
        scaling: float
            Additional scaling factor.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in1_port = self.add_input_port(image_in_tags[0])
        self.m_image_in2_port = self.add_input_port(image_in_tags[1])
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_scaling = scaling

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Add the images from the two database tags on a frame-by-frame
        basis.

        Returns
        -------
        NoneType
            None
        """

        if self.m_image_in1_port.get_shape() != self.m_image_in2_port.get_shape():
            raise ValueError("The shape of the two input tags has to be the same.")

        memory = self._m_config_port.get_attribute("MEMORY")
        nimages = self.m_image_in1_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), "Adding images...", start_time)

            images1 = self.m_image_in1_port[frames[i] : frames[i + 1],]
            images2 = self.m_image_in2_port[frames[i] : frames[i + 1],]

            self.m_image_out_port.append(
                self.m_scaling * (images1 + images2), data_dim=3
            )

        history = f"scaling = {self.m_scaling}"
        self.m_image_out_port.add_history("AddImagesModule", history)
        self.m_image_out_port.copy_attributes(self.m_image_in1_port)
        self.m_image_out_port.close_port()


class RotateImagesModule(ProcessingModule):
    """
    Pipeline module for rotating images.
    """

    __author__ = "Tomas Stolker"

    @typechecked
    def __init__(
        self, name_in: str, image_in_tag: str, image_out_tag: str, angle: float
    ) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        angle : float
            Rotation angle (deg). Rotation is clockwise for positive values.
        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_angle = angle

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Rotates all images by a constant angle.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute("MEMORY")
        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), "Rotating images...", start_time)

            images = self.m_image_in_port[frames[i] : frames[i + 1],]

            for j in range(frames[i + 1] - frames[i]):
                im_tmp = images[j,]

                # ndimage.rotate rotates in clockwise direction for positive angles
                im_tmp = rotate(im_tmp, self.m_angle, reshape=False)

                self.m_image_out_port.append(im_tmp, data_dim=3)

        history = f"angle [deg] = {self.m_angle}"
        self.m_image_out_port.add_history("RotateImagesModule", history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()


class RepeatImagesModule(ProcessingModule):
    """
    Pipeline module for repeating the images from a dataset.
    """

    __author__ = "Tomas Stolker"

    @typechecked
    def __init__(
        self, name_in: str, image_in_tag: str, image_out_tag: str, repeat: int
    ) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry with the added images that are written as output.
        repeat: int
            The number of times the input images get repeated.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_repeat = repeat

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Repeats the stack of input images a specified number of times.

        Returns
        -------
        NoneType
            None
        """

        nimages = self.m_image_in_port.get_shape()[0]
        memory = self._m_config_port.get_attribute("MEMORY")

        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i in range(self.m_repeat):
            progress(i, self.m_repeat, "Repeating images...", start_time)

            for j, _ in enumerate(frames[:-1]):
                images = self.m_image_in_port[frames[j] : frames[j + 1],]
                self.m_image_out_port.append(images, data_dim=3)

        history = f"repeat = {self.m_repeat}"
        self.m_image_out_port.add_history("RepeatImagesModule", history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.attr_writing import ParangWritingModule, AttributeWritingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestAttributeWriting:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_random(self.test_dir)
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=["parang.dat", "attribute.dat"])

    def test_input_data(self) -> None:

        data = self.pipeline.get_data("images")
        assert np.sum(data) == pytest.approx(
            0.007153603490533874, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

    def test_parang_writing(self) -> None:

        module = ParangWritingModule(
            file_name="parang.dat",
            name_in="parang_write1",
            output_dir=None,
            data_tag="images",
            header=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("parang_write1")

        data = np.loadtxt(self.test_dir + "parang.dat")
        assert np.sum(data) == pytest.approx(10.0, rel=self.limit, abs=0.0)
        assert data.shape == (5,)

    def test_attribute_writing(self) -> None:

        module = AttributeWritingModule(
            file_name="attribute.dat",
            name_in="attr_write1",
            output_dir=None,
            data_tag="images",
            attribute="PARANG",
            header=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("attr_write1")

        data = np.loadtxt(self.test_dir + "attribute.dat")

        assert np.sum(data) == pytest.approx(10.0, rel=self.limit, abs=0.0)
        assert data.shape == (5,)

    def test_attribute_not_present(self) -> None:

        module = AttributeWritingModule(
            file_name="attribute.dat",
            name_in="attr_write3",
            output_dir=None,
            data_tag="images",
            attribute="test",
            header=None,
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("attr_write3")

        assert str(error.value) == "The 'test' attribute is not present in 'images'."

    def test_parang_writing_not_present(self) -> None:

        with h5py.File(self.test_dir + "PynPoint_database.hdf5", "a") as hdf_file:
            del hdf_file["header_images/PARANG"]

        module = ParangWritingModule(
            file_name="parang.dat",
            name_in="parang_write3",
            output_dir=None,
            data_tag="images",
            header=None,
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("parang_write3")

        assert str(error.value) == "The PARANG attribute is not present in 'images'."



import os

import pytest
import numpy as np

from astropy.io import fits

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.util.tests import create_config, create_star_data, remove_test_data


class TestFitsReading:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "fits")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["fits"], files=["filenames.dat"])

    def test_fits_reading(self) -> None:

        module = FitsReadingModule(
            name_in="read1",
            input_dir=self.test_dir + "fits",
            image_tag="input",
            overwrite=False,
            check=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read1")

        data = self.pipeline.get_data("input")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_fits_reading_overwrite(self) -> None:

        module = FitsReadingModule(
            name_in="read2",
            input_dir=self.test_dir + "fits",
            image_tag="input",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read2")

        data = self.pipeline.get_data("input")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_static_not_found(self) -> None:

        self.pipeline.set_attribute("config", "DIT", "Test", static=True)

        module = FitsReadingModule(
            name_in="read3",
            input_dir=self.test_dir + "fits",
            image_tag="input",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read3")

        assert len(warning) == 2

        for item in warning:
            assert (
                item.message.args[0]
                == "Static attribute DIT (=Test) not found in the FITS "
                "header."
            )

        self.pipeline.set_attribute("config", "DIT", "ESO DET DIT", static=True)

    def test_static_changing(self) -> None:

        with fits.open(self.test_dir + "fits/images_0.fits") as hdu:
            header = hdu[0].header
            header["HIERARCH ESO DET DIT"] = 0.1
            hdu.writeto(self.test_dir + "fits/images_0.fits", overwrite=True)

        with fits.open(self.test_dir + "fits/images_1.fits") as hdu:
            header = hdu[0].header
            header["HIERARCH ESO DET DIT"] = 0.2
            hdu.writeto(self.test_dir + "fits/images_1.fits", overwrite=True)

        module = FitsReadingModule(
            name_in="read4",
            input_dir=self.test_dir + "fits",
            image_tag="input",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read4")

        assert len(warning) == 1

        assert (
            warning[0].message.args[0] == f"Static attribute ESO DET DIT has changed. "
            f"Possibly the current file {self.test_dir}fits/"
            f"images_1.fits does not belong to the data set "
            f"'input'. Attribute value is updated."
        )

    def test_header_attribute(self) -> None:

        with fits.open(self.test_dir + "fits/images_0.fits") as hdu:
            header = hdu[0].header
            header["PARANG"] = 1.0
            hdu.writeto(self.test_dir + "fits/images_0.fits", overwrite=True)

        with fits.open(self.test_dir + "fits/images_1.fits") as hdu:
            header = hdu[0].header
            header["PARANG"] = 2.0
            header["HIERARCH ESO DET DIT"] = 0.1
            hdu.writeto(self.test_dir + "fits/images_1.fits", overwrite=True)

        module = FitsReadingModule(
            name_in="read5",
            input_dir=self.test_dir + "fits",
            image_tag="input",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read5")

    def test_non_static_not_found(self) -> None:

        self.pipeline.set_attribute("config", "DIT", "None", static=True)

        for i in range(2):
            with fits.open(f"{self.test_dir}/fits/images_{i}.fits") as hdu:
                header = hdu[0].header
                del header["HIERARCH ESO DET DIT"]
                del header["HIERARCH ESO DET EXP NO"]
                hdu.writeto(f"{self.test_dir}/fits/images_{i}.fits", overwrite=True)

        module = FitsReadingModule(
            name_in="read6",
            input_dir=self.test_dir + "fits",
            image_tag="input",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read6")

        assert len(warning) == 2

        for item in warning:
            assert (
                item.message.args[0]
                == "Non-static attribute EXP_NO (=ESO DET EXP NO) not "
                "found in the FITS header."
            )

    def test_fits_read_files(self) -> None:

        module = FitsReadingModule(
            name_in="read7",
            input_dir=None,
            image_tag="files",
            overwrite=False,
            check=True,
            filenames=[
                self.test_dir + "fits/images_0.fits",
                self.test_dir + "fits/images_1.fits",
            ],
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read7")

        assert len(warning) == 2

        for item in warning:
            assert (
                item.message.args[0]
                == "Non-static attribute EXP_NO (=ESO DET EXP NO) not "
                "found in the FITS header."
            )

        data = self.pipeline.get_data("files")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_fits_read_textfile(self) -> None:

        with open(self.test_dir + "filenames.dat", "w") as file_obj:
            file_obj.write(self.test_dir + "fits/images_0.fits\n")
            file_obj.write(self.test_dir + "fits/images_1.fits\n")

        module = FitsReadingModule(
            name_in="read8",
            input_dir=None,
            image_tag="files",
            overwrite=True,
            check=True,
            filenames=self.test_dir + "filenames.dat",
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read8")

        assert len(warning) == 2

        for item in warning:
            assert (
                item.message.args[0]
                == "Non-static attribute EXP_NO (=ESO DET EXP NO) not "
                "found in the FITS header."
            )

        data = self.pipeline.get_data("files")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_fits_read_files_exists(self) -> None:

        module = FitsReadingModule(
            name_in="read9",
            input_dir=None,
            image_tag="files",
            overwrite=True,
            check=True,
            filenames=[
                f"{self.test_dir}fits/images_0.fits",
                f"{self.test_dir}fits/images_2.fits",
            ],
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("read9")

        assert (
            str(error.value)
            == f"The file {self.test_dir}fits/images_2.fits does not exist. "
            f"Please check that the path is correct."
        )

    def test_fits_read_textfile_exists(self) -> None:

        with open(self.test_dir + "filenames.dat", "w") as file_obj:
            file_obj.write(self.test_dir + "fits/images_0.fits\n")
            file_obj.write(self.test_dir + "fits/images_2.fits\n")

        module = FitsReadingModule(
            name_in="read10",
            input_dir=None,
            image_tag="files",
            overwrite=True,
            check=True,
            filenames=self.test_dir + "filenames.dat",
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("read10")

        assert (
            str(error.value)
            == f"The file {self.test_dir}fits/images_2.fits does not exist. "
            f"Please check that the path is correct."
        )



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.filter import GaussianFilterModule
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestFilter:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "data")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["data"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read",
            image_tag="data",
            input_dir=self.test_dir + "data",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)

        self.pipeline.run_module("read")

        data = self.pipeline.get_data("data")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

    def test_gaussian_filter(self) -> None:

        module = GaussianFilterModule(
            name_in="filter", image_in_tag="data", image_out_tag="filtered", fwhm=0.1
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("filter")

        data = self.pipeline.get_data("filtered")
        assert np.sum(data) == pytest.approx(
            105.54278879805275, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)



from pynpoint.core.pypeline import Pypeline

from pynpoint.processing.background import (
    SimpleBackgroundSubtractionModule,
    MeanBackgroundSubtractionModule,
    LineSubtractionModule,
    NoddingBackgroundModule,
)

from pynpoint.processing.badpixel import (
    BadPixelSigmaFilterModule,
    BadPixelInterpolationModule,
    BadPixelMapModule,
    BadPixelTimeFilterModule,
    ReplaceBadPixelsModule,
)

from pynpoint.processing.basic import (
    SubtractImagesModule,
    AddImagesModule,
    RotateImagesModule,
    RepeatImagesModule,
)

from pynpoint.processing.centering import (
    StarAlignmentModule,
    FitCenterModule,
    ShiftImagesModule,
    WaffleCenteringModule,
)

from pynpoint.processing.darkflat import DarkCalibrationModule, FlatCalibrationModule

from pynpoint.processing.extract import StarExtractionModule, ExtractBinaryModule

from pynpoint.processing.filter import GaussianFilterModule

from pynpoint.processing.fluxposition import (
    FakePlanetModule,
    SimplexMinimizationModule,
    FalsePositiveModule,
    MCMCsamplingModule,
    AperturePhotometryModule,
    SystematicErrorModule,
)

from pynpoint.processing.frameselection import (
    RemoveFramesModule,
    FrameSelectionModule,
    RemoveLastFrameModule,
    RemoveStartFramesModule,
    ImageStatisticsModule,
    FrameSimilarityModule,
    SelectByAttributeModule,
    ResidualSelectionModule,
)

from pynpoint.processing.limits import ContrastCurveModule, MassLimitsModule

from pynpoint.processing.pcabackground import (
    PCABackgroundPreparationModule,
    PCABackgroundSubtractionModule,
    DitheringBackgroundModule,
)

from pynpoint.processing.psfpreparation import (
    PSFpreparationModule,
    AngleInterpolationModule,
    AngleCalculationModule,
    SortParangModule,
    SDIpreparationModule,
)

from pynpoint.processing.psfsubtraction import (
    PcaPsfSubtractionModule,
    ClassicalADIModule,
)

from pynpoint.processing.resizing import (
    CropImagesModule,
    ScaleImagesModule,
    AddLinesModule,
    RemoveLinesModule,
)

from pynpoint.processing.stacksubset import (
    StackAndSubsetModule,
    StackCubesModule,
    DerotateAndStackModule,
    CombineTagsModule,
)

from pynpoint.processing.timedenoising import (
    CwtWaveletConfiguration,
    DwtWaveletConfiguration,
    WaveletTimeDenoisingModule,
    TimeNormalizationModule,
)

from pynpoint.readwrite.attr_reading import (
    AttributeReadingModule,
    ParangReadingModule,
    WavelengthReadingModule,
)

from pynpoint.readwrite.attr_writing import AttributeWritingModule, ParangWritingModule

from pynpoint.readwrite.fitsreading import FitsReadingModule

from pynpoint.readwrite.fitswriting import FitsWritingModule

from pynpoint.readwrite.hdf5reading import Hdf5ReadingModule

from pynpoint.readwrite.hdf5writing import Hdf5WritingModule

from pynpoint.readwrite.textwriting import TextWritingModule

from pynpoint.readwrite.nearreading import NearReadingModule

__author__ = "Tomas Stolker & Markus Bonse"
__license__ = "GPLv3"
__maintainer__ = "Tomas Stolker"
__email__ = "stolker@strw.leidenuniv.nl"



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.hdf5reading import Hdf5ReadingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestHdf5Reading:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_random(self.test_dir + "data")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["data"])

    def test_hdf5_reading(self) -> None:

        data = np.random.normal(loc=0, scale=2e-4, size=(4, 10, 10))

        with h5py.File(self.test_dir + "data/PynPoint_database.hdf5", "a") as hdf_file:
            hdf_file.create_dataset("extra", data=data)
            hdf_file.create_dataset("header_extra/PARANG", data=[1.0, 2.0, 3.0, 4.0])

        module = Hdf5ReadingModule(
            name_in="read1",
            input_filename="PynPoint_database.hdf5",
            input_dir=self.test_dir + "data",
            tag_dictionary={"images": "images"},
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read1")

        data = self.pipeline.get_data("images")
        assert np.sum(data) == pytest.approx(
            0.007153603490533874, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

    def test_dictionary_none(self) -> None:

        module = Hdf5ReadingModule(
            name_in="read2",
            input_filename="PynPoint_database.hdf5",
            input_dir=self.test_dir + "data",
            tag_dictionary=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read2")

        data = self.pipeline.get_data("images")
        assert np.sum(data) == pytest.approx(
            0.007153603490533874, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)

    def test_wrong_tag(self) -> None:

        module = Hdf5ReadingModule(
            name_in="read3",
            input_filename="PynPoint_database.hdf5",
            input_dir=self.test_dir + "data",
            tag_dictionary={"test": "test"},
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read3")

        assert len(warning) == 1
        assert (
            warning[0].message.args[0]
            == "The dataset with tag name 'test' is not found in "
            "the HDF5 file."
        )

        with h5py.File(self.test_dir + "data/PynPoint_database.hdf5", "r") as hdf_file:
            assert set(hdf_file.keys()) == set(
                ["extra", "header_extra", "header_images", "images"]
            )

    def test_no_input_filename(self) -> None:

        module = Hdf5ReadingModule(
            name_in="read4",
            input_filename=None,
            input_dir=self.test_dir + "data",
            tag_dictionary=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read4")

        data = self.pipeline.get_data("images")
        assert np.sum(data) == pytest.approx(
            0.007153603490533874, rel=self.limit, abs=0.0
        )
        assert data.shape == (5, 11, 11)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.frameselection import (
    RemoveFramesModule,
    FrameSelectionModule,
    RemoveLastFrameModule,
    RemoveStartFramesModule,
    ImageStatisticsModule,
    FrameSimilarityModule,
    SelectByAttributeModule,
    ResidualSelectionModule,
)
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestFrameSelection:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + "/"

        create_star_data(self.test_dir + "images")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["images"])

    def test_read_data(self) -> None:

        module = FitsReadingModule(
            name_in="read",
            image_tag="read",
            input_dir=self.test_dir + "images",
            overwrite=True,
            check=True,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read")

        data = self.pipeline.get_data("read")
        assert np.sum(data) == pytest.approx(
            105.54278879805277, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 11, 11)

        attr = self.pipeline.get_attribute("read", "NDIT", static=False)
        assert np.sum(attr) == pytest.approx(10, rel=self.limit, abs=0.0)
        assert attr.shape == (2,)

        attr = self.pipeline.get_attribute("read", "NFRAMES", static=False)
        assert np.sum(attr) == pytest.approx(10, rel=self.limit, abs=0.0)
        assert attr.shape == (2,)

        self.pipeline.set_attribute("read", "NDIT", [4, 4], static=False)

    def test_remove_last_frame(self) -> None:

        module = RemoveLastFrameModule(
            name_in="last", image_in_tag="read", image_out_tag="last"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("last")

        data = self.pipeline.get_data("last")
        assert np.sum(data) == pytest.approx(84.68885503527224, rel=self.limit, abs=0.0)
        assert data.shape == (8, 11, 11)

        self.pipeline.set_attribute("last", "PARANG", np.arange(8.0), static=False)
        self.pipeline.set_attribute(
            "last", "STAR_POSITION", np.full((8, 2), 5.0), static=False
        )

        attr = self.pipeline.get_attribute("last", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(28.0, rel=self.limit, abs=0.0)
        assert attr.shape == (8,)

        attr = self.pipeline.get_attribute("last", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(80.0, rel=self.limit, abs=0.0)
        assert attr.shape == (8, 2)

    def test_remove_start_frame(self) -> None:

        module = RemoveStartFramesModule(
            frames=1, name_in="start", image_in_tag="last", image_out_tag="start"
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("start")

        data = self.pipeline.get_data("start")
        assert np.sum(data) == pytest.approx(64.44307047549808, rel=self.limit, abs=0.0)
        assert data.shape == (6, 11, 11)

        attr = self.pipeline.get_attribute("start", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(24.0, rel=self.limit, abs=0.0)
        assert attr.shape == (6,)

        attr = self.pipeline.get_attribute("start", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(60.0, rel=self.limit, abs=0.0)
        assert attr.shape == (6, 2)

    def test_remove_frames(self) -> None:

        module = RemoveFramesModule(
            name_in="remove",
            image_in_tag="start",
            selected_out_tag="selected",
            removed_out_tag="removed",
            frames=[2, 5],
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("remove")

        data = self.pipeline.get_data("selected")
        assert np.sum(data) == pytest.approx(43.68337741822863, rel=self.limit, abs=0.0)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data("removed")
        assert np.sum(data) == pytest.approx(
            20.759693057269445, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 11, 11)

        attr = self.pipeline.get_attribute("selected", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(14.0, rel=self.limit, abs=0.0)
        assert attr.shape == (4,)

        attr = self.pipeline.get_attribute("selected", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(40.0, rel=self.limit, abs=0.0)
        assert attr.shape == (4, 2)

        attr = self.pipeline.get_attribute("removed", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(10.0, rel=self.limit, abs=0.0)
        assert attr.shape == (2,)

        attr = self.pipeline.get_attribute("removed", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(20.0, rel=self.limit, abs=0.0)
        assert attr.shape == (2, 2)

    def test_frame_selection(self) -> None:

        module = FrameSelectionModule(
            name_in="select1",
            image_in_tag="start",
            selected_out_tag="selected1",
            removed_out_tag="removed1",
            index_out_tag="index1",
            method="median",
            threshold=2.0,
            fwhm=0.1,
            aperture=("circular", 0.1),
            position=(None, None, 0.2),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("select1")

        data = self.pipeline.get_data("selected1")
        assert np.sum(data) == pytest.approx(54.58514780071149, rel=self.limit, abs=0.0)
        assert data.shape == (5, 11, 11)

        data = self.pipeline.get_data("removed1")
        assert np.sum(data) == pytest.approx(9.857922674786586, rel=self.limit, abs=0.0)
        assert data.shape == (1, 11, 11)

        data = self.pipeline.get_data("index1")
        assert np.sum(data) == pytest.approx(5, rel=self.limit, abs=0.0)
        assert data.shape == (1,)

        attr = self.pipeline.get_attribute("selected1", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(17.0, rel=self.limit, abs=0.0)
        assert attr.shape == (5,)

        attr = self.pipeline.get_attribute("selected1", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(50, rel=self.limit, abs=0.0)
        assert attr.shape == (5, 2)

        attr = self.pipeline.get_attribute("removed1", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(7.0, rel=self.limit, abs=0.0)
        assert attr.shape == (1,)

        attr = self.pipeline.get_attribute("removed1", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(10, rel=self.limit, abs=0.0)
        assert attr.shape == (1, 2)

        module = FrameSelectionModule(
            name_in="select2",
            image_in_tag="start",
            selected_out_tag="selected2",
            removed_out_tag="removed2",
            index_out_tag="index2",
            method="max",
            threshold=1.0,
            fwhm=0.1,
            aperture=("annulus", 0.05, 0.1),
            position=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("select2")

        data = self.pipeline.get_data("selected2")
        assert np.sum(data) == pytest.approx(21.42652724866543, rel=self.limit, abs=0.0)
        assert data.shape == (2, 11, 11)

        data = self.pipeline.get_data("removed2")
        assert np.sum(data) == pytest.approx(
            43.016543226832646, rel=self.limit, abs=0.0
        )
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data("index2")
        assert np.sum(data) == pytest.approx(10, rel=self.limit, abs=0.0)
        assert data.shape == (4,)

        attr = self.pipeline.get_attribute("selected2", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(8.0, rel=self.limit, abs=0.0)
        assert attr.shape == (2,)

        attr = self.pipeline.get_attribute("selected2", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(20, rel=self.limit, abs=0.0)
        assert attr.shape == (2, 2)

        attr = self.pipeline.get_attribute("removed2", "PARANG", static=False)
        assert np.sum(attr) == pytest.approx(16.0, rel=self.limit, abs=0.0)
        assert attr.shape == (4,)

        attr = self.pipeline.get_attribute("removed2", "STAR_POSITION", static=False)
        assert np.sum(attr) == pytest.approx(40, rel=self.limit, abs=0.0)
        assert attr.shape == (4, 2)

        module = FrameSelectionModule(
            name_in="select3",
            image_in_tag="start",
            selected_out_tag="selected3",
            removed_out_tag="removed3",
            index_out_tag="index3",
            method="range",
            threshold=(10.0, 10.7),
            fwhm=0.1,
            aperture=("circular", 0.1),
            position=None,
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("select3")

        data = self.pipeline.get_data("selected3")
        assert np.sum(data) == pytest.approx(22.2568501695632, rel=self.limit, abs=0.0)
        assert data.shape == (2, 11, 11)

        data = self.pipeline.get_data("removed3")
        assert np.sum(data) == pytest.approx(42.18622030593487, rel=self.limit, abs=0.0)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data("index3")
        assert np.sum(data) == pytest.approx(12, rel=self.limit, abs=0.0)
        assert data.shape == (4,)

    def test_image_statistics_full(self) -> None:

        module = ImageStatisticsModule(
            name_in="stat1", image_in_tag="read", stat_out_tag="stat1", position=None
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("stat1")

        data = self.pipeline.get_data("stat1")
        assert np.sum(data) == pytest.approx(
            115.68591492205017, rel=self.limit, abs=0.0
        )
        assert data.shape == (10, 6)

    def test_image_statistics_position(self) -> None:

        module = ImageStatisticsModule(
            name_in="stat2",
            image_in_tag="read",
            stat_out_tag="stat2",
            position=(5, 5, 0.1),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("stat2")

        data = self.pipeline.get_data("stat2")
        assert np.sum(data) == pytest.approx(118.7138708968444, rel=self.limit, abs=0.0)
        assert data.shape == (10, 6)

    def test_frame_similarity_mse(self) -> None:

        module = FrameSimilarityModule(
            name_in="simi1", image_tag="read", method="MSE", mask_radius=(0.0, 0.1)
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("simi1")

        attr = self.pipeline.get_attribute("read", "MSE", static=False)
        assert np.min(attr) > 0.0
        assert np.sum(attr) == pytest.approx(
            0.11739141370277852, rel=self.limit, abs=0.0
        )
        assert attr.shape == (10,)

    def test_frame_similarity_pcc(self) -> None:

        module = FrameSimilarityModule(
            name_in="simi2", image_tag="read", method="PCC", mask_radius=(0.0, 0.1)
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("simi2")

        attr = self.pipeline.get_attribute("read", "PCC", static=False)
        assert np.min(attr) > 0.0
        assert np.sum(attr) == pytest.approx(9.134820985662829, rel=self.limit, abs=0.0)
        assert attr.shape == (10,)

    def test_frame_similarity_ssim(self) -> None:

        module = FrameSimilarityModule(
            name_in="simi3",
            image_tag="read",
            method="SSIM",
            mask_radius=(0.0, 0.1),
            temporal_median="constant",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("simi3")

        attr = self.pipeline.get_attribute("read", "SSIM", static=False)
        assert np.min(attr) > 0.0
        assert np.sum(attr) == pytest.approx(9.074290801266256, rel=self.limit, abs=0.0)
        assert attr.shape == (10,)

    def test_select_by_attribute(self) -> None:

        self.pipeline.set_attribute("read", "INDEX", np.arange(44), static=False)

        module = SelectByAttributeModule(
            name_in="frame_removal_1",
            image_in_tag="read",
            attribute_tag="SSIM",
            number_frames=6,
            order="descending",
            selected_out_tag="select_sim",
            removed_out_tag="remove_sim",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("frame_removal_1")

        attr = self.pipeline.get_attribute("select_sim", "INDEX", static=False)
        assert np.sum(attr) == pytest.approx(946, rel=self.limit, abs=0.0)
        assert attr.shape == (44,)

        attr = self.pipeline.get_attribute("select_sim", "SSIM", static=False)
        assert np.sum(attr) == pytest.approx(5.545578246610884, rel=self.limit, abs=0.0)
        assert attr.shape == (6,)

        attr = self.pipeline.get_attribute("remove_sim", "SSIM", static=False)
        assert np.sum(attr) == pytest.approx(3.528712554655373, rel=self.limit, abs=0.0)
        assert attr.shape == (4,)

    def test_residual_selection(self) -> None:

        module = ResidualSelectionModule(
            name_in="residual_select",
            image_in_tag="start",
            selected_out_tag="res_selected",
            removed_out_tag="res_removed",
            percentage=80.0,
            annulus_radii=(0.1, 0.2),
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("residual_select")

        data = self.pipeline.get_data("res_selected")
        assert np.sum(data) == pytest.approx(41.77295229983322, rel=self.limit, abs=0.0)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data("res_removed")
        assert np.sum(data) == pytest.approx(
            22.670118175664847, rel=self.limit, abs=0.0
        )
        assert data.shape == (2, 11, 11)



"""
Module to obtain information about the implemented attributes.
"""

from typing import Dict, Union

from typeguard import typechecked


@typechecked
def get_attributes() -> Dict[str, Dict[str, Union[str, float, int, None]]]:
    """
    Function to get a dictionary with all attributes.

    Returns
    -------
    dict
        Attribute information.
    """

    attr = {
        "PIXSCALE": {
            "attribute": "static",
            "config": "settings",
            "value": 0.027,
            "type": "float",
        },
        "MEMORY": {
            "attribute": "static",
            "config": "settings",
            "value": 1000,
            "type": "int",
        },
        "CPU": {"attribute": "static", "config": "settings", "value": 1, "type": "int"},
        "INSTRUMENT": {
            "attribute": "static",
            "config": "header",
            "value": "INSTRUME",
            "type": "str",
        },
        "NFRAMES": {
            "attribute": "non-static",
            "config": "header",
            "value": "NAXIS3",
            "type": "int",
        },
        "EXP_NO": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO DET EXP NO",
            "type": "int",
        },
        "DIT": {
            "attribute": "static",
            "config": "header",
            "value": "ESO DET DIT",
            "type": "int",
        },
        "NDIT": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO DET NDIT",
            "type": "int",
        },
        "PARANG_START": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO ADA POSANG",
            "type": "float",
        },
        "PARANG_END": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO ADA POSANG END",
            "type": "float",
        },
        "DITHER_X": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO SEQ CUMOFFSETX",
            "type": "float",
        },
        "DITHER_Y": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO SEQ CUMOFFSETY",
            "type": "float",
        },
        "PUPIL": {
            "attribute": "non-static",
            "config": "header",
            "value": "ESO ADA PUPILPOS",
            "type": "float",
        },
        "DATE": {
            "attribute": "non-static",
            "config": "header",
            "value": "DATE-OBS",
            "type": "str",
        },
        "LATITUDE": {
            "attribute": "static",
            "config": "header",
            "value": "ESO TEL GEOLAT",
            "type": "float",
        },
        "LONGITUDE": {
            "attribute": "static",
            "config": "header",
            "value": "ESO TEL GEOLON",
            "type": "float",
        },
        "RA": {
            "attribute": "non-static",
            "config": "header",
            "value": "RA",
            "type": "float",
        },
        "DEC": {
            "attribute": "non-static",
            "config": "header",
            "value": "DEC",
            "type": "float",
        },
        "PARANG": {
            "attribute": "non-static",
            "config": "header",
            "value": "None",
            "type": "float",
        },
        "WAVELENGTH": {
            "attribute": "non-static",
            "config": "header",
            "value": "None",
            "type": "float",
        },
        "STAR_POSITION": {
            "attribute": "non-static",
            "config": None,
            "value": None,
            "type": "float",
        },
        "INDEX": {
            "attribute": "non-static",
            "config": None,
            "value": None,
            "type": "int",
        },
        "FILES": {
            "attribute": "non-static",
            "config": None,
            "value": None,
            "type": "str",
        },
    }

    return attr



import os

import pytest
import numpy as np

from astropy.io import fits

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.nearreading import NearReadingModule
from pynpoint.util.tests import create_config, create_near_data, remove_test_data


class TestNearReading:

    def setup_class(self) -> None:

        self.limit = 1e-8
        self.test_dir = os.path.dirname(__file__) + "/"
        self.fitsfile = self.test_dir + "near/images_0.fits"

        create_near_data(path=self.test_dir + "near")
        create_config(self.test_dir + "PynPoint_config.ini")

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        self.pipeline.set_attribute(
            "config", "NFRAMES", "ESO DET CHOP NCYCLES", static=True
        )
        self.pipeline.set_attribute("config", "EXP_NO", "ESO TPL EXPNO", static=True)
        self.pipeline.set_attribute("config", "NDIT", "None", static=True)
        self.pipeline.set_attribute("config", "PARANG_START", "None", static=True)
        self.pipeline.set_attribute("config", "PARANG_END", "None", static=True)
        self.pipeline.set_attribute("config", "DITHER_X", "None", static=True)
        self.pipeline.set_attribute("config", "DITHER_Y", "None", static=True)
        self.pipeline.set_attribute("config", "PIXSCALE", 0.045, static=True)
        self.pipeline.set_attribute("config", "MEMORY", 100, static=True)

        self.positions = ("chopa", "chopb")

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=["near"])

    def test_near_read(self) -> None:

        module = NearReadingModule(
            name_in="read1a",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read1a")

        for item in self.positions:

            data = self.pipeline.get_data(item)
            assert np.mean(data) == pytest.approx(0.060582854, rel=self.limit, abs=0.0)
            assert data.shape == (10, 10, 10)

    def test_near_subtract_crop_mean(self) -> None:

        module = NearReadingModule(
            name_in="read1b",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
            subtract=True,
            crop=(None, None, 0.3),
            combine="mean",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read1b")

        data = self.pipeline.get_data(self.positions[0])
        assert np.mean(data) == pytest.approx(0.0, rel=self.limit, abs=0.0)
        assert data.shape == (2, 7, 7)

        data = self.pipeline.get_data(self.positions[1])
        assert np.mean(data) == pytest.approx(0.0, rel=self.limit, abs=0.0)
        assert data.shape == (2, 7, 7)

    def test_near_median(self) -> None:

        module = NearReadingModule(
            name_in="read1c",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
            combine="median",
        )

        self.pipeline.add_module(module)
        self.pipeline.run_module("read1c")

        data = self.pipeline.get_data(self.positions[0])
        assert np.mean(data) == pytest.approx(0.060582854, rel=self.limit, abs=0.0)
        assert data.shape == (2, 10, 10)

        data = self.pipeline.get_data(self.positions[1])
        assert np.mean(data) == pytest.approx(0.060582854, rel=self.limit, abs=0.0)
        assert data.shape == (2, 10, 10)

    def test_static_not_found(self) -> None:

        self.pipeline.set_attribute("config", "DIT", "Test", static=True)

        module = NearReadingModule(
            name_in="read2",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read2")

        assert len(warning) == 4

        for item in warning:
            assert (
                item.message.args[0]
                == "Static attribute DIT (=Test) not found in the FITS "
                "header."
            )

        self.pipeline.set_attribute("config", "DIT", "ESO DET SEQ1 DIT", static=True)

    def test_nonstatic_not_found(self) -> None:

        self.pipeline.set_attribute("config", "NDIT", "Test", static=True)

        module = NearReadingModule(
            name_in="read3",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read3")

        assert len(warning) == 4

        for item in warning:
            assert (
                item.message.args[0]
                == "Non-static attribute NDIT (=Test) not found in the "
                "FITS header."
            )

        self.pipeline.set_attribute("config", "NDIT", "None", static=True)

    def test_check_header(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            hdulist[0].header["ESO DET CHOP ST"] = "F"
            hdulist[0].header["ESO DET CHOP CYCSKIP"] = 1
            hdulist[0].header["ESO DET CHOP CYCSUM"] = "T"
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(
            name_in="read4",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read4")

        assert len(warning) == 3

        assert warning[0].message.args[0] == "Dataset was obtained without chopping."
        assert warning[1].message.args[0] == "Chop cycles (1) have been skipped."
        assert warning[2].message.args[0] == "FITS file contains averaged images."

        with fits.open(self.fitsfile) as hdulist:
            hdulist[0].header["ESO DET CHOP ST"] = "T"
            hdulist[0].header["ESO DET CHOP CYCSKIP"] = 0
            hdulist[0].header["ESO DET CHOP CYCSUM"] = "F"
            hdulist.writeto(self.fitsfile, overwrite=True)

    def test_frame_type_invalid(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            hdulist[10].header["ESO DET FRAM TYPE"] = "Test"
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(
            name_in="read5",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("read5")

        assert (
            str(error.value)
            == "Frame type (Test) not a valid value. Expecting HCYCLE1 or "
            "HCYCLE2 as value for ESO DET FRAM TYPE."
        )

    def test_frame_type_missing(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            hdulist[10].header.remove("ESO DET FRAM TYPE")
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(
            name_in="read6",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module("read6")

        assert (
            str(error.value)
            == "Frame type not found in the FITS header. Image number: 9."
        )

    def test_same_cycle(self) -> None:

        with fits.open(self.fitsfile) as hdulist:

            with pytest.warns(UserWarning) as warning:
                hdulist[10].header["ESO DET FRAM TYPE"] = "HCYCLE1"

            assert len(warning) == 1

            assert (
                warning[0].message.args[0]
                == "Keyword name 'ESO DET FRAM TYPE' is greater "
                "than 8 characters or contains characters not "
                "allowed by the FITS standard; a HIERARCH card "
                "will be created."
            )

            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(
            name_in="read7",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read7")

        assert len(warning) == 2

        assert (
            warning[0].message.args[0]
            == "Previous and current chop position (HCYCLE1) are "
            "the same. Skipping the current image."
        )

        assert (
            warning[1].message.args[0]
            == "The number of images is not equal for chop A and "
            "chop B."
        )

    def test_odd_number_images(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            del hdulist[11]
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(
            name_in="read8",
            input_dir=self.test_dir + "near",
            chopa_out_tag=self.positions[0],
            chopb_out_tag=self.positions[1],
        )

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module("read8")

        assert len(warning) == 2

        assert (
            warning[0].message.args[0] == f"FITS file contains odd number of images: "
            f"{self.fitsfile}"
        )

        assert (
            warning[1].message.args[0]
            == "The number of chop cycles (5) is not equal to "
            "half the number of available HDU images (4)."
        )
